#include "network.h"

#define IF_NUM_MAX 50
int interface_count;
struct interface_context interfaces[IF_NUM_MAX];
int max_socket = 0;

unsigned char lldp_multicast_address[6] = {
	0x01, 0x80, 0xc2, 0x00, 0x00, 0x0e};

extern int verbose;

/*******************************************************/
/*                         init                        */
/*******************************************************/
int open_socket(struct interface_context *ifc);
int bind_socket(struct interface_context *ifc);
int determine_local_mac(struct interface_context *ifc);
int init_dcbx_specific_fields(struct interface_context *ifc);

int reinit_interface(struct interface_context *ifc)
{
	int ret;
	if ((ret = open_socket(ifc)) < 0) return ret;
	if ((ret = bind_socket(ifc)) < 0) return ret;
	if ((ret = determine_local_mac(ifc)) < 0) return ret;
	if ((ret = init_dcbx_specific_fields(ifc)) < 0) return ret;
	return 0;
}

int init_interfaces()
{
	int ret, i;
	for (i = 0; i < interface_count; i++) {
		ret = reinit_interface(interfaces + i);
		if (ret < 0) return ret;
	}
	return 0;
}

int open_socket(struct interface_context *ifc)
{
	struct ifreq ifr;
	int rc;

	ifc->socket_handle = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_DCBX));
	if (max_socket < ifc->socket_handle)
		max_socket = ifc->socket_handle;

	memset(&(ifc->sll), 0, sizeof(ifc->sll));
	ifc->sll.sll_family = AF_PACKET;
	
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifc->name, sizeof (ifr.ifr_name));
	rc = ioctl(ifc->socket_handle, SIOCGIFINDEX, &ifr);
	if (rc == -1) {
		printf("open_socket %s: error: %s\n", ifc->name, strerror(errno));
		return -1;
	}
	ifc->sll.sll_ifindex = ifc->if_index = ifr.ifr_ifindex;	
	return 0;
}

int bind_socket(struct interface_context *ifc)
{
	int i, rc;
	struct packet_mreq pm;

	ifc->sll.sll_family = AF_PACKET;
	ifc->sll.sll_protocol = htons(ETH_P_DCBX);
	
	rc = bind(ifc->socket_handle, (struct sockaddr*)&(ifc->sll), sizeof(struct sockaddr_ll));
        if (rc == -1) {
		printf("bind error %s: %s\n", ifc->name, strerror(errno));
		return -1;
        }
	
	memset(&pm, 0 ,sizeof(struct packet_mreq));
		
	pm.mr_ifindex = ifc->sll.sll_ifindex;
	pm.mr_type = PACKET_MR_MULTICAST;

	for(i = 0; i < 6; i++)
		pm.mr_address[i] = lldp_multicast_address[i];
	pm.mr_alen = 6;

	rc = setsockopt(ifc->socket_handle, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &pm, sizeof(pm));
	if (rc == -1) {
		printf("setsockopt error %s: %s\n", ifc->name, strerror(errno));
		return -1;
	}
	return 0;
}

int determine_local_mac(struct interface_context *ifc)
{
	int i, rc;
	struct ifreq ifr;

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifc->name, sizeof (ifr.ifr_name));
	rc = ioctl(ifc->socket_handle, SIOCGIFHWADDR, &ifr);
	if (rc == -1) {
		printf("socket ioctl error %s: %s\n", ifc->name, strerror(errno));
		return -1;
	}

	for (i = 0; i < 6; i++) 
		ifc->mac[i] = ifr.ifr_hwaddr.sa_data[i];
	return 0;
}

/*******************************************************/
/*                       read/send                     */
/*******************************************************/
void print_packet(int is_incoming, unsigned char mac[6], struct interface_context *ifc,unsigned char* buf, int size);

void send_buffer_to(unsigned char to[6], struct interface_context *ifc, unsigned char *buf, int length)
{
        int bytes_sent;	
	int i;

	for (i = 0; i < 6; i++)
		ifc->sll.sll_addr[i] = to[i];
	if (verbose > 1) print_packet(0, to, ifc, buf, length);
        bytes_sent = sendto(ifc->socket_handle, buf, length, 0, (struct sockaddr*)&(ifc->sll), sizeof(struct sockaddr_ll));
        if(bytes_sent < 0)
                printf("Error sending packet in %s: %s\n", ifc->name, strerror(errno) );
}

void send_buffer(struct interface_context *ifc, unsigned char *buf, int length)
{
	send_buffer_to(lldp_multicast_address, ifc, buf, length);
}

void read_from_interface(struct interface_context *ifc, 
			 void (*handle_packet)(unsigned char from[6], struct interface_context *ifc,unsigned char* buf, int size))
{	
	unsigned char buf[1000];	
	socklen_t slen;
	int i, size;
	unsigned char from[6];

	slen = sizeof(struct sockaddr_ll);
	size = recvfrom(ifc->socket_handle, (void *)buf, sizeof(buf), 0, (struct sockaddr*)&(ifc->sll), &slen);	
	if (size < 0) {
		printf("rceive from socket error in %s: %s\n", ifc->name, strerror(errno));
	} else if (size > 0) {
		assert((ifc->sll.sll_protocol & 0xff) == 0x88);
		assert(((ifc->sll.sll_protocol & 0xff00) >> 8) == 0xcc);
		assert(ifc->sll.sll_ifindex == ifc->if_index);


		for (i = 0; i < 6; i++)
			from[i]  = ifc->sll.sll_addr[i];

		if (verbose > 1) print_packet(1, from, ifc, buf, size);
		handle_packet(from, ifc, buf, size);
	}
}

/*******************************************************/
/*                    multiplexing                     */
/*******************************************************/
void select_dispatcher(
	void (*handle_packet)(unsigned char from[6], struct interface_context *ifc, unsigned char* buf, int size), 
	void (*handle_error)(struct interface_context *ifc))
{
	fd_set rfds_err;
	fd_set rfds_in;
	struct timeval tv;
	int ret;
	int i;

	FD_ZERO(&rfds_err);
	for (i = 0; i < interface_count; i++)
		FD_SET((unsigned int)interfaces[i].socket_handle, &rfds_err);
	
	FD_ZERO(&rfds_in);
	for (i = 0; i < interface_count; i++)
		FD_SET((unsigned int)interfaces[i].socket_handle, &rfds_in);

	/* Wait up to 1 seconds. */
	tv.tv_sec = 1;
	tv.tv_usec = 0;

	ret = select(max_socket + 1, &rfds_in, NULL, &rfds_err, &tv);

	if (ret > 0) {
		for (i = 0; i < interface_count; i++) {
			if (FD_ISSET(interfaces[i].socket_handle, &rfds_err)) {			
				printf("error in select from socket in %s\n", interfaces[i].name);
				handle_error(interfaces + i);
			}
			if (FD_ISSET(interfaces[i].socket_handle, &rfds_in)) {			
				read_from_interface(&(interfaces[i]), handle_packet);
			}
		}
		
	}
	if (ret < 0) 
		printf("select error: %s\n", strerror(errno));
}

void recv_dcbx (
	void (*handle_packet)(unsigned char from[6], struct interface_context *ifc,unsigned char* buf, int size),
		void (*handle_error)(struct interface_context *ifc))
{
	time_t now, start;

	time(&start);	
	do {
		select_dispatcher(handle_packet, handle_error);
		time(&now);
	} while(now < start + 1);
}

int find_interface_by_name(char *name)
{
	int i;

	for (i = 0; i < interface_count; i++)
		if (strncmp(interfaces[i].name, name, sizeof(name)) == 0)
			return i;
	return -1;
}

int create_vlan(struct interface_context *ifc)
{
	struct vlan_ioctl_args ifr;
	struct ifreq ifr2;

	if (verbose) printf("creating %s.0 with egress 0:%d\n", ifc->name, ifc->prio);

	memset(&ifr, 0, sizeof(struct vlan_ioctl_args));
	ifr.cmd = ADD_VLAN_CMD;
	ifr.u.VID = 0;
	memcpy (ifr.device1, ifc->name, strlen(ifc->name));

	if (ioctl(ifc->socket_handle, SIOCSIFVLAN, &ifr) < 0) {
		//printf("vlan ioctl error: %s\n", strerror(errno));
		//returns error when virtual interface already exists.
	}	

	memset(&ifr, 0, sizeof(struct vlan_ioctl_args));
	ifr.cmd = SET_VLAN_EGRESS_PRIORITY_CMD;
	ifr.u.skb_priority = 0;

	ifr.vlan_qos = ifc->prio;
	sprintf(ifr.device1, "%s.%d", ifc->name, 0);

	if (ioctl(ifc->socket_handle, SIOCSIFVLAN, &ifr) < 0) {
		printf("vlan ioctl error: %s\n", strerror(errno));
	}

	sprintf(ifr2.ifr_name, "%s.%d", ifc->name, 0);
	if (ioctl(ifc->socket_handle, SIOCGIFFLAGS, &ifr2) < 0) {
		printf("set interface up ioctl error: %s\n", strerror(errno));
	}

	ifr2.ifr_flags |= IFF_UP;

	if (ioctl(ifc->socket_handle, SIOCSIFFLAGS, &ifr2) < 0) {
		printf("set interface up ioctl error: %s\n", strerror(errno));
	}

	return 0;
}


