/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>

#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/qp.h>
#include <linux/mlx4/cq.h>

#include "mlx4_fc.h"

static void mlx4_fc_arm_cq(struct mlx4_fc_cq *cq)
{
	mlx4_cq_arm(&cq->mcq, MLX4_CQ_DB_REQ_NOT,
		    cq->vhba->fport->mfcdev->uar_map,
		    MLX4_GET_DOORBELL_LOCK(&cq->vhba->fport->mfcdev->uar_lock));
}

static void mlx4_fc_cq_event(struct mlx4_cq *cq, enum mlx4_event event)
{
	FCPRINT("%s has been called. event = 0x%x\n", __FUNCTION__,
		(unsigned int)event);
}

void mlx4_fc_cq_clean(struct mlx4_fc_cq *cq)
{
	struct mlx4_cq *mcq = &cq->mcq;
	struct mlx4_fc_vhba *vhba = cq->vhba;
	struct mlx4_cqe *cqe, cqe2;

	cqe = (struct mlx4_cqe *)cq->buf + (mcq->cons_index & cq->size_mask);

	while (XNOR
	       (cqe->owner_sr_opcode & MLX4_CQE_OWNER_MASK,
		mcq->cons_index & cq->size)) {
		if ((cqe->owner_sr_opcode & MLX4_CQE_OPCODE_MASK) == 0x1e) {
			struct mlx4_err_cqe
			    __attribute__ ((__unused__)) * ecqe =
			    (struct mlx4_err_cqe *)cqe;

			FCPRINT
			    ("Error completion. qpn: 0x%08x, wqe index: 0x%08x, vendor error syndrome: 0x%x, syndrome: 0x%x\n",
			     be32_to_cpu(ecqe->my_qpn),
			     be16_to_cpu(ecqe->wqe_index),
			     ecqe->vendor_err_syndrome, ecqe->syndrome);
			return;
		}

		cqe2 = *cqe;
		mcq->cons_index++;
		mlx4_cq_set_ci(mcq);

		if (cqe->owner_sr_opcode & MLX4_CQE_IS_SEND_MASK) {
			ASSERT_PR(cq->comp_tx,
				  "Shouldn't have completion on %s TX",
				  cq->name);
			cq->comp_tx(vhba, &cqe2);
		} else {
			ASSERT_PR(cq->comp_rx,
				  "Shouldn't have completion on %s RX",
				  cq->name);
			cq->comp_rx(vhba, &cqe2);
		}

		cqe =
		    (struct mlx4_cqe *)cq->buf +
		    (mcq->cons_index & cq->size_mask);
	}
}

static void mlx4_fc_cq_comp(struct mlx4_cq *mcq)
{
	struct mlx4_fc_cq *cq = container_of(mcq, struct mlx4_fc_cq, mcq);

	mlx4_fc_cq_clean(cq);
	mlx4_fc_arm_cq(cq);
}

int mlx4_fc_create_cq(struct mlx4_fc_vhba *vhba, struct mlx4_fc_cq *cq,
		      int entries, int eqidx, int arm, comp_fn comp_rx,
		      comp_fn comp_tx, char *name)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	int err;

	cq->vhba = vhba;
	cq->comp_rx = comp_rx;
	cq->comp_tx = comp_tx;
	strncpy(cq->name, name, sizeof(cq->name));

	cq->size = roundup_pow_of_two(entries + 1);
	cq->size_mask = cq->size - 1;
	cq->buf_size = cq->size * sizeof(struct mlx4_cqe);

	err =
	    mlx4_alloc_hwq_res(mfcdev->dev, &cq->wqres, cq->buf_size,
			       cq->buf_size);
	if (err)
		return err;

	cq->mcq.set_ci_db = cq->wqres.db.db;
	cq->mcq.arm_db = cq->wqres.db.db + 1;
	*cq->mcq.set_ci_db = 0;
	*cq->mcq.arm_db = 0;

	cq->buf = (struct mlx4_fc_cqe *)cq->wqres.buf.direct.buf;

	err = mlx4_cq_alloc(mfcdev->dev, cq->size, &cq->wqres.mtt,
			    &mfcdev->priv_uar, cq->wqres.db.dma, &cq->mcq,
			    eqidx, 0);
	if (err)
		goto err_man;

	cq->mcq.comp = mlx4_fc_cq_comp;
	cq->mcq.event = mlx4_fc_cq_event;

	if (arm)
		mlx4_fc_arm_cq(cq);

	return 0;

err_man:
	mlx4_free_hwq_res(mfcdev->dev, &cq->wqres, cq->buf_size);
	return err;
}

void mlx4_fc_destroy_cq(struct mlx4_fc_cq *cq)
{
	struct mlx4_fc_dev *mfcdev = cq->vhba->fport->mfcdev;

	mlx4_cq_free(mfcdev->dev, &cq->mcq);
	mlx4_free_hwq_res(mfcdev->dev, &cq->wqres, cq->buf_size);
	cq->buf_size = 0;
	cq->buf = NULL;
}
