/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/log2.h>
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_fc.h>
#include <scsi/scsi_tcq.h>
#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/cq.h>
#include <scsi/fc_frame.h>
#include <scsi/fc/fc_fcp.h>
#include <scsi/fc/fc_els.h>
#include <scsi/fc/fc_fcoe.h>
#include <scsi/libfc.h>

#include "mlx4_fc.h"
#include "bitmap.h"

static enum mlx4_fcmd_io_dir scsi_dir_translate(enum dma_data_direction dmadir)
{
	switch (dmadir) {
	case DMA_BIDIRECTIONAL:
		return FCMD_IO_DIR_BIDI;
	case DMA_FROM_DEVICE:
		return FCMD_IO_DIR_READ;
	case DMA_TO_DEVICE:
		return FCMD_IO_DIR_WRITE;
	case DMA_NONE:
		return FCMD_IO_DIR_TARGET;	/* also used for initiator when no RDMA expected */
	default:
		FCPRINT("unknown DMA dir: %d\n", dmadir);
	}
	return -1;
}

static void mlx4_fc_fcmd_tx_comp(struct mlx4_fc_vhba *vhba,
				 struct mlx4_cqe *cqe)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_q *sq = &vhba->fcmd.fc_qp.sq;
	struct mlx4_fc_data_seg *data;
	struct mlx4_fc_fcmd_tx_desc *tx_desc;
	u64 dma;
	u32 count;
	unsigned long flags;
	int wqe_idx = be16_to_cpu(cqe->wqe_index) & sq->size_mask;

	tx_desc = sq->buf + wqe_idx * FCMD_SQ_BB_SIZE;

	data = &tx_desc->data;
	dma = be64_to_cpu(data->addr);
	count = be32_to_cpu(data->count);

	pci_unmap_single(mfcdev->dev->pdev, dma, count, PCI_DMA_TODEVICE);

	spin_lock_irqsave(&sq->lock, flags);
	sq->cons++;
	spin_unlock_irqrestore(&sq->lock, flags);
}

static int mlx4_fc_map_sg_to_fmr(struct mlx4_fc_dev *mfcdev,
				 struct scatterlist *sglist, int nents,
				 struct mlx4_fc_fexch *fexch,
				 enum dma_data_direction dir)
{
	struct mlx4_fmr *fmr = &fexch->fmr;
	struct scatterlist *sg;
	int page_cnt, sg_cnt;
	unsigned int total_len;
	int i;
	u64 fmr_page_mask = SHIFT_TO_MASK(fmr->page_shift);	/* masks off low bits */
	u64 dma;
	u64 page_list[MLX4_FC_MAX_FMR_PAGES];
	unsigned int fmr_page_size = SHIFT_TO_SIZE(fmr->page_shift);
	u32 rkey, lkey;
	int rc = 0;
	int err;

#if DEBUG_SG
	char buf[1024];
	char *p = buf;
#endif

	sg_cnt = pci_map_sg(mfcdev->dev->pdev, sglist, nents, dir);
	ASSERT(sg_cnt > 0);

#if (!USE_FBO)
	/* if we can't use FBO, we exit here if first dma is not alligned */
	if (sg_dma_address(sglist) & ~fmr_page_mask) {
		rc = -EAGAIN;
		goto out_unmap;
	}
#endif

#if DEBUG_SG
	p += sprintf(p, "FMR_MAP: fmr_pg_sz: 0x%x sg_cnt=%d fbo=0x%llx ",
		     SHIFT_TO_SIZE(fmr->page_shift), sg_cnt,
		     sglist & ~fmr_page_mask);
#endif

	page_cnt = 0;
	total_len = 0;
	for_each_sg(sglist, sg, sg_cnt, i) {
		total_len += sg_dma_len(sg);

#if DEBUG_SG
		p += sprintf(p, "sg[%d]={0x%llx, 0x%x} ", i, sg_dma_address(sg),
			     dma_len);
#endif

		if (sg_dma_address(sg) & ~fmr_page_mask) {
			if (i > 0) {
				FCLOG(fc_log_data,
				      "Non-first page is partial, will retry with smaller granularity FMR\n");
				rc = -EINVAL;
				goto out_unmap;
			}
		}
		if ((sg_dma_address(sg) + sg_dma_len(sg)) & ~fmr_page_mask) {
			if (i < sg_cnt - 1) {
				FCLOG(fc_log_data,
				      "Non-last page is partial, will retry with smaller granularity FMR\n");
				rc = -EINVAL;
				goto out_unmap;
			}
		}

		for (dma = (sg_dma_address(sg) & fmr_page_mask);
		     dma < sg_dma_address(sg) + sg_dma_len(sg);
		     dma += fmr_page_size) {
			ASSERT(page_cnt < fmr->max_pages);
#if DEBUG_SG
			p += sprintf(p, "pg[%d]=0x%llx ", page_cnt, dma);
#endif

			page_list[page_cnt] = dma;
			++page_cnt;
		}
	}

#if DEBUG_SG
	p += sprintf(p, "| len=0x%x", total_len);
	ASSERT(p < buf + sizeof(buf));
	FCLOG(fc_log_data, "%s\n", buf);
#endif

	ASSERT(page_cnt);
	err = mlx4_map_phys_fmr_fbo(mfcdev->dev, fmr, page_list, page_cnt,	/* num_pages */
				    0,	/* iova */
				    sg_dma_address(sglist) & ~fmr_page_mask, total_len, &lkey, &rkey,	/* fake, don't need them */
				    1	/* use same key */
	    );

	if (err) {
		mlx4_err(mfcdev->dev, "Could not map FMR to given SG\n");
		rc = err;
		goto out_unmap;
	}

	return 0;

out_unmap:
	pci_unmap_sg(mfcdev->dev->pdev, sglist, nents, dir);
	return rc;
}

static int mlx4_fc_create_bounce(struct mlx4_fc_dev *mfcdev,
				 struct scsi_cmnd *cmd,
				 struct mlx4_fc_fexch *fexch)
{
	void *bouncebuff;
	int bounceoff = fexch->bounce_off;
	unsigned int total_len;
	struct scatterlist *sg;
	void *page_addr;

	int i;

	bouncebuff = kmalloc(BOUNCESIZE, GFP_ATOMIC);
	if (!bouncebuff)
		return -ENOMEM;
#if DEBUG_BOUNCE
	memset(bouncebuff, MLX_MAGIC, BOUNCESIZE);
#endif
	ASSERT(scsi_bufflen(cmd) < BOUNCESIZE - bounceoff);

	total_len = 0;
	scsi_for_each_sg(cmd, sg, scsi_sg_count(cmd), i) {
		// if write, copy data from orig sg to bounce buff
		if (cmd->sc_data_direction == DMA_TO_DEVICE) {
			page_addr = kmap_atomic(sg_page(sg), KM_SOFTIRQ0);
			memcpy(bouncebuff + bounceoff + total_len,
			       page_addr + (sg->offset & ~PAGE_MASK),
			       sg->length);
			kunmap_atomic(page_addr, KM_SOFTIRQ0);
		}
		total_len += sg->length;
	}
	ASSERT_PR(scsi_bufflen(cmd) == total_len,
		  "request_bufflen=%d, total len in sg=%d\n",
		  scsi_bufflen(cmd), total_len);

	sg_init_one(&fexch->bounce_sg[0], bouncebuff + bounceoff,
		    scsi_bufflen(cmd));
	fexch->bouncebuff = bouncebuff;

	return 0;
}

static int mlx4_fc_map_fmr(struct mlx4_fc_dev *mfcdev, struct scsi_cmnd *cmd,
			   struct mlx4_fc_fexch *fexch)
{
	int err;

	if (cmd->sc_data_direction == DMA_NONE)
		return 0;

	ASSERT(cmd->sc_data_direction != DMA_BIDIRECTIONAL);
	if (cmd->sc_data_direction == DMA_BIDIRECTIONAL) {
		mlx4_warn(mfcdev->dev, "BIDI SCSI commands not supported\n");
		return -EINVAL;
	}

	ASSERT(scsi_sg_count(cmd));

	fexch->bouncebuff = NULL;
	fexch->bounce_off = 0;
	err = mlx4_fc_map_sg_to_fmr(mfcdev,
				    scsi_sglist(cmd), scsi_sg_count(cmd),
				    fexch, cmd->sc_data_direction);
	if (err == -EAGAIN) {
		err = mlx4_fc_create_bounce(mfcdev, cmd, fexch);
		if (err)
			return err;
		err = mlx4_fc_map_sg_to_fmr(mfcdev, &fexch->bounce_sg[0], 1,
					    fexch, cmd->sc_data_direction);
	}

	return err;
}

static void mlx4_fc_unmap_fmr_sg(struct mlx4_fc_dev *mfcdev,
				 struct scatterlist *sglist, int nents,
				 struct mlx4_fmr *fmr,
				 enum dma_data_direction dir)
{
	u32 dummy_lkey, dummy_rkey;

	pci_unmap_sg(mfcdev->dev->pdev, sglist, nents, dir);
	mlx4_fmr_unmap(mfcdev->dev, fmr, &dummy_lkey, &dummy_rkey);
}

static void mlx4_fc_destroy_bounce(struct mlx4_fc_dev *mfcdev,
				   struct scsi_cmnd *cmd,
				   struct mlx4_fc_fexch *fexch)
{
	struct scatterlist *sg;
	int i;
	unsigned long total_len;
	char *page_addr;

	if (cmd->sc_data_direction == DMA_FROM_DEVICE) {
#if DEBUG_BOUNCE
		/* check write before start of buffer (as given to fmr) */
		for (p = fexch->bouncebuff;
		     p < fexch->bouncebuff + fexch->bounce_off; ++p)
			ASSERT_PR(*p == MLX_MAGIC,
				  "WRITE BEFORE START OF BUFFER\n"
				  "  buff=%p (dma=0x%llx), bounceoff=0x%x\n"
				  "  at p=%p magic removed, instead got:\n"
				  "  [p-2]=0x%02x [p-1]=0x%02x [p]=0x%02x [p+1]=0x%02x [p+2]=0x%02x [p+3]=0x%02x\n",
				  fexch->bouncebuff, fexch->bounce_dma,
				  fexch->bounce_off, p, *(p - 2), *(p - 1), *p,
				  *(p + 1), *(p + 2), *(p + 3));

		/* check write beyond buffer len (as given to fmr) */
		for (p =
		     fexch->bouncebuff + fexch->bounce_off +
		     cmd->request_bufflen;
		     p < fexch->bouncebuff + fexch->bounce_off; ++p)
			ASSERT_PR(*p == MLX_MAGIC,
				  "WRITE BEYOND END OF BUFFER\n"
				  "  buff=%p (dma=0x%llx), bounceoff=0x%x, request_buflen=0x%x\n"
				  "  at p=%p magic removed, instead got:\n"
				  "  [p-2]=0x%02x [p-1]=0x%02x [p]=0x%02x [p+1]=0x%02x [p+2]=0x%02x [p+3]=0x%02x\n",
				  fexch->bouncebuff, fexch->bounce_dma,
				  fexch->bounce_off, cmd->request_bufflen, p,
				  *(p - 2), *(p - 1), *p, *(p + 1), *(p + 2),
				  *(p + 3));
#endif

		total_len = 0;
		scsi_for_each_sg(cmd, sg, scsi_sg_count(cmd), i) {
			page_addr = kmap_atomic(sg_page(sg), KM_SOFTIRQ0);
			memcpy(page_addr + (sg->offset & ~PAGE_MASK),
			       fexch->bouncebuff + fexch->bounce_off +
			       total_len, sg->length);
			kunmap_atomic(page_addr, KM_SOFTIRQ0);
			total_len += sg->length;
		}
		ASSERT(scsi_bufflen(cmd) == total_len);
	}

	kfree(fexch->bouncebuff);
	fexch->bouncebuff = NULL;
}

static void mlx4_fc_unmap_fmr(struct mlx4_fc_dev *mfcdev, struct scsi_cmnd *cmd,
			      struct mlx4_fc_fexch *fexch)
{
	if (cmd->sc_data_direction == DMA_NONE)
		return;

	if (cmd->sc_data_direction == DMA_BIDIRECTIONAL) {
		mlx4_warn(mfcdev->dev, "BIDI SCSI commands not supported\n");
		return;
	}

	if (fexch->bouncebuff) {
		mlx4_fc_unmap_fmr_sg(mfcdev, &fexch->bounce_sg[0], 1,
				     &fexch->fmr, cmd->sc_data_direction);
		mlx4_fc_destroy_bounce(mfcdev, cmd, fexch);
	} else
		mlx4_fc_unmap_fmr_sg(mfcdev, scsi_sglist(cmd),
				     scsi_sg_count(cmd), &fexch->fmr,
				     cmd->sc_data_direction);
}

void mlx4_fc_clean_exch(struct mlx4_fc_fexch *fexch)
{
	struct mlx4_fc_vhba *vhba = fexch->vhba;
	int xno = fexch - vhba->fexch;

	mlx4_fc_unmap_fmr(vhba->fport->mfcdev, fexch->sc_cmd, fexch);
	mlx4_fc_bitmap_slot_free(&vhba->fexch_bm, xno);
}

/*
 * FEXCH completion - pay attention: ethernet header is stripped.
 */
static void mlx4_fc_fexch_rx_comp(struct mlx4_fc_vhba *vhba,
				  struct mlx4_cqe *generic_cqe)
{
	struct mlx4_fc_fexch_cqe *cqe = (struct mlx4_fc_fexch_cqe *)generic_cqe;
	struct mlx4_fc_fexch *fexch;
	struct mlx4_fc_q *rq;
	int wqe_idx;
	struct mlx4_fc_rx_desc *rx_desc;
	int xno;
	u32 qpn;
	unsigned long flags;
	struct fcp_resp *fc_rp;
	size_t rxcnt;

	qpn = be32_to_cpu(cqe->my_qpn) & ((1 << 24) - 1);
	xno = qpn - vhba->base_fexch_qpn;
	fexch = &vhba->fexch[xno];

	rq = &fexch->fc_qp.rq;

	del_timer /*_sync*/ (&fexch->timeout_timer);
	if (fexch->timed_out)
		goto out;

	wqe_idx = be16_to_cpu(cqe->wqe_index) & rq->size_mask;
	rx_desc = rq->buf + (wqe_idx * rq->stride);

	pci_unmap_single(vhba->fport->mfcdev->dev->pdev,
			 be64_to_cpu(rx_desc->data[0].addr),
			 be32_to_cpu(rx_desc->data[0].count),
			 PCI_DMA_FROMDEVICE);

	FCLOG(fc_log_data, "FCMD response: xno=0x%x, qpn=0x%x\n", xno, qpn);

#ifdef DEBUG_CPU_BINDING
	if (xno % num_online_cpus() != smp_processor_id())
		FCPRINT
		    ("SMP: got completion for xno %d on cpu %d instead of %d\n",
		     xno, smp_processor_id(), xno % num_online_cpus());
#endif

	rxcnt = be32_to_cpu(cqe->rcv_byte_count);

	fc_rp = (struct fcp_resp *)(fexch->response_buf + 2 + 2 + 24);	// ETHER_TYPE(2) + FCOE(2) + FC(24)

	mlx4_fc_rx_scsi(vhba, fc_rp, fexch->sc_cmd, rxcnt, &fexch->fmr, xno);

	mlx4_fc_clean_exch(fexch);

      out:
	spin_lock_irqsave(&rq->lock, flags);
	rq->cons++;
	spin_unlock_irqrestore(&rq->lock, flags);

}

static void mlx4_fc_fexch_timeout_work(struct work_struct *work)
{
	struct mlx4_fc_fexch *fexch =
	    container_of(work, struct mlx4_fc_fexch, timeout_work);
	struct mlx4_fc_vhba *vhba = fexch->vhba;
	struct scsi_cmnd *sc_cmd = fexch->sc_cmd;

	ASSERT(vhba);
	ASSERT(sc_cmd);

	FCPRINT("Aborting fexch[0x%lx]\n", fexch - vhba->fexch);

	mlx4_fc_clean_exch(fexch);

	sc_cmd->result = (DID_BUS_BUSY << 16);

	sc_cmd->scsi_done(sc_cmd);

	/* TODO: do fexch reset */
}

static void mlx4_fc_fexch_timeout(unsigned long context)
{
	struct mlx4_fc_fexch *fexch = (struct mlx4_fc_fexch *)context;

	fexch->timed_out = 1;

	FCPRINT("FEXCH timeout called\n");

	INIT_WORK(&fexch->timeout_work, mlx4_fc_fexch_timeout_work);
	queue_work(mlx4_fc_fcoe_wq, &fexch->timeout_work);
}

static void mlx4_fc_qp_event(struct mlx4_qp *qp, enum mlx4_event type)
{
	FCPRINT("qp event for qpn=0x%08x event_type=0x%x\n", qp->qpn, type);
}

static int mlx4_fc_create_fexch(struct mlx4_fc_vhba *vhba, int xno)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_fexch *fexch = &vhba->fexch[xno];
	struct mlx4_fc_qp *qp = &fexch->fc_qp;
	struct mlx4_fc_q *rq = &qp->rq;
	int err = 0;

	fexch->vhba = vhba;

	mlx4_fc_q_init(rq, FEXCH_RQ_WQE_SIZE, FEXCH_RQ_NUM_WQES, 0);

	qp->buf_size = (FEXCH_SQ_BB_SIZE * FEXCH_SQ_NUM_BBS) +
		       (rq->size * rq->stride);

	err = mlx4_alloc_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size,
				 qp->buf_size);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate man for FEXCH %d\n",
			 xno);
		goto err_free_rxinfo;
	}

	rq->buf = qp->wqres.buf.direct.buf +
		  (FEXCH_SQ_BB_SIZE * FEXCH_SQ_NUM_BBS);

	*qp->wqres.db.db = 0;

	err = mlx4_qp_alloc(mfcdev->dev, vhba->base_fexch_qpn + xno, &qp->mqp);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate QP number 0x%x\n",
			 qp->mqp.qpn);
		goto err_free_man;
	}

	qp->doorbell_qpn = swab32(qp->mqp.qpn << 8);
        qp->mqp.event = mlx4_fc_qp_event;

	ASSERT(MLX4_FC_FMR_PAGE_SHIFT >=
	       (ffs(mfcdev->dev->caps.page_size_cap) - 1));
	err = mlx4_fmr_alloc_reserved(mfcdev->dev, vhba->base_fexch_mpt + xno,
				      mfcdev->priv_pdn | MLX4_MPT_ENABLE_INVALIDATE,
				      MLX4_PERM_REMOTE_WRITE | MLX4_PERM_REMOTE_READ,
				      MLX4_FC_MAX_FMR_PAGES, 1 /* max_maps */,
				      MLX4_FC_FMR_PAGE_SHIFT, &fexch->fmr);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "Could not allocate FMR for fexch %d, err=%d\n", xno,
			 err);
		goto err_free_qp;
	}

	err = mlx4_fmr_enable(mfcdev->dev, &fexch->fmr);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "Could not enable FMR for fexch %d, err=%d\n", xno,
			 err);
		goto err_free_fmr;
	}

	return 0;

err_free_fmr:
	mlx4_fmr_free_reserved(mfcdev->dev, &fexch->fmr);
err_free_qp:
	mlx4_qp_remove(mfcdev->dev, &fexch->fc_qp.mqp);
	mlx4_qp_free(mfcdev->dev, &fexch->fc_qp.mqp);
err_free_man:
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
err_free_rxinfo:
	mlx4_fc_q_destroy(rq);
	return err;
}

static void mlx4_fc_destroy_fexch(struct mlx4_fc_vhba *vhba, int xno)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_fexch *fexch = &vhba->fexch[xno];
	struct mlx4_fc_qp *qp = &fexch->fc_qp;
	struct mlx4_fc_q *rq = &qp->rq;

	mlx4_fmr_free_reserved(mfcdev->dev, &fexch->fmr);
	if (qp->is_created)
		mlx4_qp_to_reset(mfcdev->dev, &qp->mqp);
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
	mlx4_fc_q_destroy(rq);
}

int mlx4_fc_init_fexch(struct mlx4_fc_vhba *vhba, int xno)
{
	struct mlx4_fc_dev *mfcdev = vhba->fport->mfcdev;
	struct mlx4_fc_fexch *fexch = &vhba->fexch[xno];
	struct mlx4_fc_qp *qp = &fexch->fc_qp;
	enum mlx4_qp_state qp_state = MLX4_QP_STATE_RST;
	int err = 0;

	struct mlx4_qp_context context = {
		.flags = cpu_to_be32(QPC_SERVICE_TYPE_FEXCH << 16),
		.pd = cpu_to_be32(mfcdev->priv_pdn),
		/* Raw-ETH requirement */
		.mtu_msgmax = 0x77,
		/* this means SQ_NUM_BBS=1, and SQ_BB_SIZE=1 */
		.sq_size_stride = 0,
		.rq_size_stride =
		    ilog2(FEXCH_RQ_NUM_WQES) << 3 | ilog2(FEXCH_RQ_WQE_SIZE >>
							  4),
		.usr_page = cpu_to_be32(mfcdev->priv_uar.index),
		.local_qpn = cpu_to_be32(qp->mqp.qpn),
		.pri_path.sched_queue =
		    0x83 | vhba->fc_vlan_prio << 3 | (vhba->fport->port -
						      1) << 6,
		.pri_path.counter_index = 0xff,
		.pri_path.ackto = MLX4_LINK_TYPE_ETH,
		/* Source MAC index */
		.pri_path.grh_mylmc = vhba->rfci[RFCI_DATA].fc_mac_idx,
		.params2 = cpu_to_be32((qp->wqres.buf.direct.map &
					(PAGE_SIZE - 1)) & 0xfc0),
		.cqn_send =
		    cpu_to_be32(vhba->fexch_cq[xno % num_online_cpus()].mcq.
				cqn),
		.cqn_recv =
		    cpu_to_be32(vhba->fexch_cq[xno % num_online_cpus()].mcq.
				cqn),
		.db_rec_addr = cpu_to_be64(qp->wqres.db.dma),
		.srqn = 0,
		.my_fc_id_idx = 0,
	};

	if (vhba->fc_vlan_idx != -1) {
		context.pri_path.fl = 0x40;
		context.pri_path.snooper_flags = vhba->fc_vlan_idx;
	}

	err = mlx4_qp_to_ready(mfcdev->dev, &qp->wqres.mtt, &context, &qp->mqp,
			       &qp_state);

	if (qp_state != MLX4_QP_STATE_RST)
		qp->is_created = 1;

	if (qp_state != MLX4_QP_STATE_RTS) {
		mlx4_err(mfcdev->dev,
			 "Error bringing FEXCH %d QP to RTS state, qpn=0x%x\n",
			 xno, qp->mqp.qpn);
		return err;
	}
	return 0;
}

/*
 * Attention: This function could be called from interrupt context
 */
int mlx4_fc_create_fcmd(struct mlx4_fc_vhba *vhba)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_qp *qp = &vhba->fcmd.fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	int err = 0;
	int i, eqidx, cpu;

	err = mlx4_fc_q_init(sq, FCMD_SQ_BB_SIZE, vhba->num_fexch,
			     sizeof(struct fcp_cmnd *));
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not alloc info for fcmd sq\n",
			 fc_port->port, vhba->idx);
		goto err_out;
	}

	for (i = 0; i < sq->size; i++) {
		struct fcp_cmnd *cdb_cmd;

		cdb_cmd = mlx4_fc_q_info_get(sq, i, struct fcp_cmnd *) =
		    kmalloc(sizeof(struct fcp_cmnd), GFP_KERNEL);
		if (!cdb_cmd) {
			goto err_free_txinfo;
		}
		memset(cdb_cmd, 0, sizeof(*cdb_cmd));
	}

	vhba->fcmd.fc_qp.buf_size =
	    (sq->size * sq->stride) + (FCMD_RQ_NUM_WQES * FCMD_RQ_WQE_SIZE);

	err =
	    mlx4_alloc_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size,
			       qp->buf_size);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not allocate man for fcmd, err=%d\n",
			 fc_port->port, vhba->idx, err);
		goto err_free_txinfo;
	}

	sq->buf = qp->wqres.buf.direct.buf;
	*qp->wqres.db.db = 0;

	mlx4_fc_stamp_sq(sq);

	err = mlx4_qp_reserve_range(mfcdev->dev, 1, 1, &qp->mqp.qpn);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not reserve QPN for fcmd, err=%d\n",
			 fc_port->port, vhba->idx, err);
		goto err_free_man;
	}

	err = mlx4_qp_alloc(mfcdev->dev, qp->mqp.qpn, &qp->mqp);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not allocate QPN 0x%x\n",
			 fc_port->port, vhba->idx, qp->mqp.qpn);
		goto err_release_qp;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: FCMD QPN=0x%x\n", fc_port->port,
	      vhba->idx, qp->mqp.qpn);

	qp->doorbell_qpn = swab32(qp->mqp.qpn << 8);

	err = mlx4_fc_create_cq(vhba, &vhba->fcmd.fc_cq,
				MLX4_FC_FCMD_CQ_ENTRIES,
				MLX4_LEAST_ATTACHED_VECTOR,/* don't mind which cpu */
				0,	/* don't arm the cq */
				NULL, mlx4_fc_fcmd_tx_comp, "FCMD");
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Failed creating FCMD CQ, err=%d\n",
			 fc_port->port, vhba->idx, err);
		goto err_free_qp;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: FCMD CQN=0x%x\n", fc_port->port,
	      vhba->idx, vhba->fcmd.fc_cq.mcq.cqn);

	/* Create FEXCHs for this FCMD */
	vhba->fexch = vmalloc(vhba->num_fexch * sizeof(struct mlx4_fc_fexch));
	if (!vhba->fexch) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Error allocating fexch array\n",
			 fc_port->port, vhba->idx);
		goto err_free_cq;

	}
	memset(vhba->fexch, 0, vhba->num_fexch * sizeof(struct mlx4_fc_fexch));
	for (i = 0; i < vhba->num_fexch; i++) {
		vhba->fexch[i].response_buf =
		    kmalloc(MLX4_FC_FCMD_RX_SKB_BUFSIZE, GFP_KERNEL);
		if (!vhba->fexch[i].response_buf) {
			mlx4_err(mfcdev->dev,
				 "port%d vhba%d fexch %d: Error allocating fexch response buffer\n",
				 fc_port->port, vhba->idx, i);
			goto err_free_fexch_arr;
		}
	}

	err = mlx4_fc_bitmap_alloc(&vhba->fexch_bm, vhba->num_fexch);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Error allocating fexch bitmap for\n",
			 fc_port->port, vhba->idx);
		goto err_free_fexch_arr;
	}

	eqidx = 0;
	for_each_online_cpu(cpu) {	/* TODO: No need for per-cpu */
		err = mlx4_fc_create_cq(vhba, &vhba->fexch_cq[eqidx],
					vhba->num_fexch / num_online_cpus(),
					(eqidx % num_online_cpus()) %
					 mfcdev->dev->caps.num_comp_vectors,
					1,/* arm cq */
					mlx4_fc_fexch_rx_comp, NULL, "FEXCH");
		if (err) {
			mlx4_err(mfcdev->dev,
				 "port%d vhba%d: Failed creating FEXCH CQ %d, err=%d\n",
				 fc_port->port, vhba->idx, eqidx, err);
			goto err_destroy_fexch_cq;
		}
		FCLOG(fc_log_ctrl,
		      "port%d vhba%d: Created FEXCH CQ %d, CQN=0x%x\n",
		      fc_port->port, vhba->idx, eqidx,
		      vhba->fexch_cq[eqidx].mcq.cqn);

		++eqidx;
	}

	for (i = 0; i < vhba->num_fexch; i++) {
		err = mlx4_fc_create_fexch(vhba, i);
		if (err) {
			mlx4_err(mfcdev->dev,
				 "port%d vhba%d: Failed creating FEXCH %d of %d, err=%d\n",
				 fc_port->port, vhba->idx, i, vhba->num_fexch,
				 err);
			goto err_destroy_fexch;
		}
	}

	FCLOG(fc_log_ctrl, "port%d vhba%d: Created %d FEXCHs, QPN=0x%x-0x%x\n",
	      fc_port->port, vhba->idx, i,
	      vhba->fexch[0].fc_qp.mqp.qpn, vhba->fexch[i - 1].fc_qp.mqp.qpn);

	return 0;

err_destroy_fexch:
	while (--i >= 0)
		mlx4_fc_destroy_fexch(vhba, i);
err_destroy_fexch_cq:
	while (--eqidx >= 0)
		mlx4_fc_destroy_cq(&vhba->fexch_cq[eqidx]);
	mlx4_fc_bitmap_free(&vhba->fexch_bm);
err_free_fexch_arr:
	for (i = 0; i < vhba->num_fexch; i++) {
		if (!vhba->fexch[i].response_buf)
			break;
		kfree(vhba->fexch[i].response_buf);
	}
	vfree(vhba->fexch);
err_free_cq:
	mlx4_fc_destroy_cq(&vhba->fcmd.fc_cq);
err_free_qp:
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
err_release_qp:
	mlx4_qp_release_range(mfcdev->dev, qp->mqp.qpn, 1);
err_free_man:
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
err_free_txinfo:
	for (i = 0; i < sq->size; i++) {
		if (!mlx4_fc_q_info_get(sq, i, struct fcp_cmnd *))
			 break;
		kfree(mlx4_fc_q_info_get(sq, i, struct fcp_cmnd *));
	}
	mlx4_fc_q_destroy(sq);
err_out:
	return err;
}

void mlx4_fc_destroy_fcmd(struct mlx4_fc_vhba *vhba)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_qp *qp = &vhba->fcmd.fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	int i;

	for (i = 0; i < vhba->num_fexch; ++i)
		mlx4_fc_destroy_fexch(vhba, i);
	for (i = 0; i < num_online_cpus(); ++i)
		mlx4_fc_destroy_cq(&vhba->fexch_cq[i]);
	ASSERT(mlx4_fc_bitmap_empty(&vhba->fexch_bm));
	mlx4_fc_bitmap_free(&vhba->fexch_bm);

	for (i = 0; i < vhba->num_fexch; i++) {
		if (!vhba->fexch[i].response_buf)
			break;
		kfree(vhba->fexch[i].response_buf);
	}
	vfree(vhba->fexch);

	mlx4_fc_destroy_cq(&vhba->fcmd.fc_cq);
	if (qp->is_created)
		mlx4_qp_to_reset(mfcdev->dev, &qp->mqp);
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
	mlx4_qp_release_range(mfcdev->dev, vhba->fcmd.fc_qp.mqp.qpn, 1);
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);

	for (i = 0; i < sq->size; i++) {
		if (!mlx4_fc_q_info_get(sq, i, struct fcp_cmnd *))
			 break;
		kfree(mlx4_fc_q_info_get(sq, i, struct fcp_cmnd *));
	}
	mlx4_fc_q_destroy(sq);
}

int mlx4_fc_init_fcmd(struct mlx4_fc_vhba *vhba)
{
	struct mlx4_fc_dev *mfcdev = vhba->fport->mfcdev;
	struct mlx4_fc_qp *qp = &vhba->fcmd.fc_qp;
	enum mlx4_qp_state qp_state = MLX4_QP_STATE_RST;
	int err = 0;
	int i;

	struct mlx4_qp_context context = {
		.flags = cpu_to_be32(QPC_SERVICE_TYPE_FCMD << 16),
		.pd = cpu_to_be32(mfcdev->priv_pdn),
		.mtu_msgmax = 0x77,
		.sq_size_stride =
		    ilog2(vhba->
			  num_fexch) << 3 | ilog2(FCMD_SQ_BB_SIZE >> 4) |
		    SQ_NO_PREFETCH,
		/* this means RQ_NUM_WQES=1, and RQ_WQE_SIZE=1 */
		.rq_size_stride = 0,
		.usr_page = cpu_to_be32(mfcdev->priv_uar.index),
		.local_qpn = cpu_to_be32(qp->mqp.qpn),
		.pri_path.sched_queue =
		    0x83 | vhba->fc_vlan_prio << 3 | (vhba->fport->port -
						      1) << 6,
		.pri_path.counter_index = 0xff,
		.pri_path.ackto = MLX4_LINK_TYPE_ETH,
		.pri_path.grh_mylmc = vhba->rfci[RFCI_DATA].fc_mac_idx,
		.params2 =
		    cpu_to_be32((qp->wqres.buf.direct.
				 map & (PAGE_SIZE - 1)) & 0xfc0),
		.cqn_send = cpu_to_be32(vhba->fcmd.fc_cq.mcq.cqn),
		.cqn_recv = cpu_to_be32(vhba->fcmd.fc_cq.mcq.cqn),
		.db_rec_addr = cpu_to_be64(qp->wqres.db.dma),
		.srqn = 0,
		.VE = 0,
		.exch_base = cpu_to_be16(vhba->base_fexch_qpn),
		.exch_size = ilog2(vhba->num_fexch),
		.my_fc_id_idx = 0,
	};
	if (vhba->fc_vlan_idx != -1) {
		context.pri_path.fl = 0x40;
		context.pri_path.snooper_flags = vhba->fc_vlan_idx;
		FCLOG(fc_log_ctrl, "FCMD set to use vlan idx %d\n",
		      vhba->fc_vlan_idx);
	}

	err = mlx4_qp_to_ready(mfcdev->dev, &qp->wqres.mtt, &context, &qp->mqp,
			       &qp_state);

	if (qp_state != MLX4_QP_STATE_RST)
		qp->is_created = 1;

	if (qp_state != MLX4_QP_STATE_RTS) {
		mlx4_err(mfcdev->dev, "Error bringing FCMD QP to RTS state\n");
		return err;
	}

	/* bring FEXCHs to ready state */
	for (i = 0; i < vhba->num_fexch; i++) {
		err = mlx4_fc_init_fexch(vhba, i);
		if (err) {
			mlx4_err(mfcdev->dev,
				 "Failed init of FEXCH %d for vhba, err=%d\n",
				 i, err);
			return err;
		}
	}
	FCLOG(fc_log_ctrl, "Initialized %d FEXCHs for vhba\n", i);

	return 0;
}

static inline void set_ctrl_seg(struct mlx4_fc_ctrl_seg *ctrl, u8 seqid,
				u8 info, u8 ls, u32 task_retry_id)
{
	ctrl->size = cpu_to_be16(((FCMD_SQ_BB_SIZE / 16) & 0x3f) | (1 << 7));
	ctrl->flags = cpu_to_be32(MLX4_FC_BIT_TX_COMP |	/* request completion */
				  (seqid << 24) | (info << 20) | (ls << 16));
	ctrl->parameter = cpu_to_be32(task_retry_id);
}

static inline int prepare_fexch(struct mlx4_fc_vhba *vhba,
				struct scsi_cmnd *sc_cmd)
{
	struct mlx4_fc_fexch *fexch;
	struct mlx4_fc_dev *mfcdev = vhba->fport->mfcdev;
	int fexch_idx;
	int err;
	int index;

	fexch_idx = mlx4_fc_bitmap_slot_alloc(&vhba->fexch_bm);
	if (fexch_idx == -1) {
		mlx4_err(mfcdev->dev, "No free FEXCH\n");
		return -1;
	}
	fexch = &vhba->fexch[fexch_idx];

	err = mlx4_fc_map_fmr(mfcdev, sc_cmd, fexch);
	if (err == -EAGAIN) {
		FCPRINT("Could not map %dB resolution FMR. exch no %d\n",
			(1 << MLX4_FC_FMR_PAGE_SHIFT), fexch_idx);
	}
	ASSERT_PR(err == 0, "xno=%d\n", fexch_idx);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not map SCSI sg to FMR\n");
		return -1;
	}

	index = mlx4_fc_post_rx_buf(mfcdev, &fexch->fc_qp,
				    fexch->response_buf,
				    MLX4_FC_FCMD_RX_SKB_BUFSIZE);
	ASSERT_PR(index > -1, "could not post a single recv buff on fexch %d\n",
		  fexch_idx);

	fexch->sc_cmd = sc_cmd;
	init_timer(&fexch->timeout_timer);
	fexch->timeout_timer.data = (unsigned long)fexch;
	fexch->timeout_timer.function = mlx4_fc_fexch_timeout;
	fexch->timeout_timer.expires = jiffies + MLX4_FC_FCMD_TIMEOUT;
	add_timer(&fexch->timeout_timer);

	mlx4_fc_ring_db_rx(&fexch->fc_qp);

	return fexch_idx;
}

static inline void set_init_seg(struct mlx4_fc_init_seg *init, int frame_size,
				u32 remote_fid,
				enum dma_data_direction data_dir, int fexch_idx)
{
	enum mlx4_fcmd_io_dir mlx_dir = scsi_dir_translate(data_dir);

	ASSERT(mlx_dir >= 0);
	ASSERT(mlx_dir != FCMD_IO_DIR_BIDI);

	init->pe = 0;		/* priority enable, goes to F_CTL[17] */
	init->cs_ctl = 0;	/* CS_CTL/Priority field */
	init->seq_id_tx = 0;	/* seq. id to be used in FCP_DATA frames */
	init->mtu = cpu_to_be16(frame_size / 4);
	init->remote_fid[2] = (remote_fid) & 0xff;
	init->remote_fid[1] = (remote_fid >> 8) & 0xff;
	init->remote_fid[0] = (remote_fid >> 16) & 0xff;

	init->flags = (1 << 1) |	/* originator, 1=initiator */
		      (scsi_dir_translate(data_dir) << 3) |
		      (0x0 << 6); /* abort code, goes to F_CTL[5:4] */

	/* initiators never know remote exch no. at beginning of exch */
	init->remote_exch = cpu_to_be16(0xffff);
	/* alloc free exchange, put index here */
	init->local_exch_idx = cpu_to_be16(fexch_idx);
}

DEFINE_SPINLOCK(cq_poll);

int mlx4_do_fcmd_send(struct scsi_cmnd *sc_cmd,
		      void (*done) (struct scsi_cmnd *))
{
	struct fc_lport *lp;
	struct mlx4_fc_vhba *vhba;
	struct mlx4_fc_dev *mfcdev;
	struct mlx4_dev *mdev;
	struct mlx4_fc_q *sq;
	int fexch_idx;
	struct fc_rport *rport;
	struct mlx4_fc_fcmd_tx_desc *tx_desc;
	struct mlx4_fc_data_seg *data;
	dma_addr_t dma;
	struct fcp_cmnd *cdb_cmd;
	u32 index, prod;
	__be32 op_own;
	unsigned long flags;

	lp = shost_priv(sc_cmd->device->host);
	FCLOG(fc_log_data, "scsi command. sc_cmd = %p\n", sc_cmd);

	vhba = lport_priv(lp);
	mfcdev = vhba->fport->mfcdev;
	mdev = mfcdev->dev;
	sq = &vhba->fcmd.fc_qp.sq;
	rport = starget_to_rport(scsi_target(sc_cmd->device));

	sc_cmd->scsi_done = done;

	if (spin_trylock_irqsave(&cq_poll, flags)) {
		mlx4_fc_cq_clean(&vhba->fcmd.fc_cq);
		spin_unlock_irqrestore(&cq_poll, flags);
	}

	/* Check available SQ BBs + 1 spare SQ BB for owenership */
	spin_lock_irqsave(&sq->lock, flags);
	if (unlikely((u32) (sq->prod - sq->cons - 1) > sq->size - 2)) {
		mlx4_info(mdev, "port%d vhba%d: Out of FCMD send queue BBs\n",
			  vhba->fport->port, vhba->idx);
		spin_unlock_irqrestore(&sq->lock, flags);
		return -1;
	}
	prod = sq->prod;
	++sq->prod;
	spin_unlock_irqrestore(&sq->lock, flags);

	index = prod & sq->size_mask;
	cdb_cmd = mlx4_fc_q_info_get(sq, index, struct fcp_cmnd *);

	tx_desc = sq->buf + index * FCMD_SQ_BB_SIZE;

	/* set ctrl segment: for initiator we must have info=6
	   and ls=0. seqid and task_retry_id set to 0 */
	set_ctrl_seg(&tx_desc->ctrl, 0, 6, 0, 0);

	/* allocate and prepare FEXCH for command */
	fexch_idx = prepare_fexch(vhba, sc_cmd);
	if (fexch_idx < 0)
		return -1;

	/* set init segment */
	set_init_seg(&tx_desc->init, rport->maxframe_size,
		     rport->port_id, sc_cmd->sc_data_direction,
		     fexch_idx);

	/* set addr segment */
	tx_desc->addr.static_rate = 0;
	memcpy(tx_desc->addr.dmac, vhba->dest_addr, ETH_ALEN);

	/* prepare cdb command in buffer */
	if (sc_cmd->sc_data_direction == DMA_FROM_DEVICE) {
		cdb_cmd->fc_flags = FCP_CFL_RDDATA;
	} else if (sc_cmd->sc_data_direction == DMA_TO_DEVICE) {
		cdb_cmd->fc_flags = FCP_CFL_WRDATA;
	} else {
		cdb_cmd->fc_flags = 0;
	}

	cdb_cmd->fc_dl = htonl(scsi_bufflen(sc_cmd));
	cdb_cmd->fc_flags &= ~FCP_CFL_LEN_MASK;
	int_to_scsilun(sc_cmd->device->lun, (struct scsi_lun *)cdb_cmd->fc_lun);

	memcpy(cdb_cmd->fc_cdb, sc_cmd->cmnd, sc_cmd->cmd_len);

	/* set data segment */
	dma =
	    pci_map_single(mfcdev->dev->pdev, cdb_cmd, sizeof(*cdb_cmd),
			   PCI_DMA_TODEVICE);
	data = &tx_desc->data;
	data->addr = cpu_to_be64(dma);
	data->count = cpu_to_be32(sizeof(*cdb_cmd));
	data->mem_type = cpu_to_be32(mfcdev->mr.key);	/* always snoop */

	op_own = cpu_to_be32(MLX4_FC_FCMD_OP_SEND) |
	    ((prod & sq->size) ? cpu_to_be32(MLX4_FC_BIT_DESC_OWN) : 0);

	/* Ensure new descirptor (and ownership of next descirptor) hits memory
	 * before setting ownership of this descriptor to HW */
	wmb();
	tx_desc->ctrl.op_own = op_own;

	FCLOG_MAC(fc_log_data,
		  "FCMD to FID 0x%06x, max_payload=%d, exch_idx=0x%x, "
		  "fexch_qpn=0x%x, WQ_idx=0x%x, scsi reqlen=%d, scsi_dir=%d, dest_MAC=",
		  tx_desc->addr.dmac, rport->port_id, rport->maxframe_size,
		  fexch_idx, vhba->fexch[fexch_idx].fc_qp.mqp.qpn, index,
		  scsi_bufflen(sc_cmd), sc_cmd->sc_data_direction);

	/* Ring doorbell! */
	wmb();
	writel(vhba->fcmd.fc_qp.doorbell_qpn,
	       mfcdev->uar_map + MLX4_SEND_DOORBELL);

	return 0;
}

/* High level stuff */
