/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/if_vlan.h>
#include <linux/kobject.h>

#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/qp.h>
#include <linux/mlx4/device.h>
#include <scsi/libfc.h>
#include <linux/kthread.h>

#include "mlx4_fc.h"

MODULE_AUTHOR(MLX4FC_DRIVER_AUTHOR);
MODULE_DESCRIPTION(MLX4FC_DRIVER_DESC);
MODULE_LICENSE(MLX4FC_DRIVER_LICENSE);
MODULE_VERSION(MLX4FC_DRIVER_VERSION);

#if FC_DEBUG
int mlx4_fc_log_level = MLX4_FC_DEFAULT_LOG_LEVEL;
module_param_named(log_level, mlx4_fc_log_level, int, 0644);
MODULE_PARM_DESC(log_level,
		 "Debug level for logging. 0=Off, 1=CTRL path, 2=DATA path."
		 "Default =  " __stringify(MLX4_FC_DEFAULT_LOG_LEVEL));
#endif

int mlx4_debug_mode = 0;
module_param_named(debug_mode, mlx4_debug_mode, int, 0644);
MODULE_PARM_DESC(debug_mode,
		 " 0 = None (default), 1 = use gw_mac as dest and burnt MAC as src.");

char *mlx4_gw_mac;
module_param_named(gw_mac, mlx4_gw_mac, charp, 0644);
MODULE_PARM_DESC(gw_mac,
		 "GW MAC. Used for Debug Mode 1. Format: XX:XX:XX:XX:XX:XX");
uint8_t gw_mac[ETH_ALEN] = { 0, 0, 0, 0, 0, 0 };

static const char mlx4_fc_version[] __devinitdata =
    MLX4FC_DRIVER_NAME ": " MLX4FC_DRIVER_DESC " v"
    MLX4FC_DRIVER_VERSION " (" MLX4FC_DRIVER_RELDATE ")\n";

LIST_HEAD(mfcdev_list);
DEFINE_SPINLOCK(mfcdev_list_lock);

struct workqueue_struct *mlx4_fc_fcoe_wq;
struct scsi_transport_template *mlx4_fc_transport_template;

static int mlx4_fc_reset(struct Scsi_Host *shost);

struct fc_function_template mlx4_fc_transport_function = {
	.show_host_node_name = 1,
	.show_host_port_name = 1,
	.show_host_supported_classes = 1,
	.show_host_supported_fc4s = 1,
	.show_host_active_fc4s = 1,
	.show_host_maxframe_size = 1,

	.show_host_port_id = 1,
	.show_host_supported_speeds = 1,
	.get_host_speed = fc_get_host_speed,
	.show_host_speed = 1,
	.show_host_port_type = 1,
	.get_host_port_state = fc_get_host_port_state,
	.show_host_port_state = 1,
	.show_host_symbolic_name = 1,

	.dd_fcrport_size = sizeof(struct fc_rport_libfc_priv),
	.show_rport_maxframe_size = 1,
	.show_rport_supported_classes = 1,

	.show_host_fabric_name = 1,
	.show_starget_node_name = 1,
	.show_starget_port_name = 1,
	.show_starget_port_id = 1,
	.set_rport_dev_loss_tmo = fc_set_rport_loss_tmo,
	.show_rport_dev_loss_tmo = 1,
	.get_fc_host_stats = fc_get_host_stats,
	.issue_fc_host_lip = mlx4_fc_reset,

	.terminate_rport_io = fc_rport_terminate_io,
};

static int mlx4_fc_reset(struct Scsi_Host *shost)
{
	struct fc_lport *lp = shost_priv(shost);
	fc_lport_reset(lp);
	return 0;
}

int mlx4_fc_q_init(struct mlx4_fc_q *q, u16 stride, size_t size,
		   size_t info_size)
{
	q->prod = 0;
	q->cons = 0xffffffff;
	q->stride = stride;
	q->size = size;
	q->size_mask = q->size - 1;
	q->info = NULL;

	if (info_size) {
		q->info = (u8 *) vmalloc(q->size * info_size);
		if (!q->info) {
			return -ENOMEM;
		}
	}

	spin_lock_init(&q->lock);
	memset(q->info, 0, q->size * info_size);

	return 0;
}

void mlx4_fc_q_destroy(struct mlx4_fc_q *q)
{
	if (!q->info)
		return;

	vfree(q->info);
	q->info = NULL;
}

void mlx4_fc_stamp_sq(struct mlx4_fc_q *sq)
{
	__be32 *p;
	int i;

	/* stamp first dword of every 64 byte */
	for (i = 0; i < sq->size; ++i) {
		p = sq->buf + i * sq->stride;
		*p = cpu_to_be32(1 << 31);
	}

}

static void *mlx4_fc_add(struct mlx4_dev *mdev)
{
	struct mlx4_fc_dev *mfcdev;
	int port;
	int err;
	unsigned long flags;

	FCLOG(fc_log_ctrl, "%.*s: mlx4 device added\n", MLX4_BOARD_ID_LEN,
	      mdev->board_id);

	mfcdev = kzalloc(sizeof(struct mlx4_fc_dev), GFP_KERNEL);
	if (!mfcdev) {
		dev_err(&mdev->pdev->dev, "Device struct alloc failed, "
			"aborting.\n");
		err = -ENOMEM;
		goto err;
	}

	if (mlx4_pd_alloc(mdev, &mfcdev->priv_pdn))
		goto err_free_dev;
	FCLOG(fc_log_ctrl, "%.*s: PDN: 0x%x\n", MLX4_BOARD_ID_LEN,
	      mdev->board_id, mfcdev->priv_pdn);

	err = mlx4_mr_alloc(mdev, mfcdev->priv_pdn, 0, ~0ull,
			    MLX4_PERM_LOCAL_WRITE | MLX4_PERM_LOCAL_READ, 0, 0,
			    &mfcdev->mr);

	if (err) {
		dev_err(&mdev->pdev->dev, "mr alloc failed");
		err = -ENOMEM;
		goto err_free_pd;
	}

	err = mlx4_mr_enable(mdev, &mfcdev->mr);
	if (err) {
		dev_err(&mdev->pdev->dev, "mr enable failed");
		err = -ENOMEM;
		goto err_free_mr;
	}
	FCLOG(fc_log_ctrl, "%.*s: Memory key is 0x%x\n", MLX4_BOARD_ID_LEN,
	      mdev->board_id, mfcdev->mr.key);

	if (mlx4_uar_alloc(mdev, &mfcdev->priv_uar))
		goto err_free_mr;

	mfcdev->uar_map =
	    ioremap(mfcdev->priv_uar.pfn << PAGE_SHIFT, PAGE_SIZE);
	if (!mfcdev->uar_map)
		goto err_free_uar;
	MLX4_INIT_DOORBELL_LOCK(&mfcdev->uar_lock);

	INIT_LIST_HEAD(&mfcdev->pgdir_list);
	mutex_init(&mfcdev->pgdir_mutex);

	mfcdev->dev = mdev;
	mfcdev->dma_device = &mdev->pdev->dev;
	mfcdev->num_fc_ports = 0;
	mlx4_foreach_port(port, mfcdev->dev, MLX4_PORT_TYPE_ETH) {
		mfcdev->num_fc_ports++;
		FCLOG(fc_log_ctrl, "%.*s: port %d is ETH ports\n",
		      MLX4_BOARD_ID_LEN, mdev->board_id, port);
	}

	mlx4_query_steer_cap(mdev, &mfcdev->log_num_mac, &mfcdev->log_num_vlan,
			     &mfcdev->log_num_prio);
	mlx4_qp_get_region(mdev, MLX4_QP_REGION_FC_ADDR, &mfcdev->base_rfci_qpn,
			   &mfcdev->num_rfci_qps);

	FCLOG(fc_log_ctrl, "%.*s: num_macs=%d, num_vlans=%d, num_prios=%d, "
	      "base_rfci_qpn=0x%x, rfci_cnt=0x%x\n",
	      MLX4_BOARD_ID_LEN, mdev->board_id,
	      mfcdev->log_num_mac, mfcdev->log_num_vlan, mfcdev->log_num_prio,
	      mfcdev->base_rfci_qpn, mfcdev->num_rfci_qps);

	mlx4_foreach_port(port, mfcdev->dev, MLX4_PORT_TYPE_ETH) {
		err = mlx4_fc_init_port(mfcdev, port);
		if (err)
			goto err_free_ports;
	}

	spin_lock_irqsave(&mfcdev_list_lock, flags);
	list_add(&mfcdev->list, &mfcdev_list);
	spin_unlock_irqrestore(&mfcdev_list_lock, flags);

	return mfcdev;

err_free_ports:
	while (--port)
		mlx4_fc_free_port(mfcdev, port);
	iounmap(mfcdev->uar_map);
err_free_uar:
	mlx4_uar_free(mdev, &mfcdev->priv_uar);
err_free_mr:
	mlx4_mr_free(mfcdev->dev, &mfcdev->mr);
err_free_pd:
	mlx4_pd_free(mdev, mfcdev->priv_pdn);
err_free_dev:
	kfree(mfcdev);
err:
	return NULL;
}

static void mlx4_fc_remove(struct mlx4_dev *dev, void *fcdev_ptr)
{
	struct mlx4_fc_dev *mfcdev = fcdev_ptr;
	int port;
	unsigned long flags;

	FCLOG(fc_log_ctrl, "%.*s: removing\n", MLX4_BOARD_ID_LEN,
	      dev->board_id);
	spin_lock_irqsave(&mfcdev_list_lock, flags);
	list_del(&mfcdev->list);
	spin_unlock_irqrestore(&mfcdev_list_lock, flags);

	mlx4_foreach_port(port, mfcdev->dev, MLX4_PORT_TYPE_ETH) {
		mlx4_fc_free_port(mfcdev, port);
	}
	iounmap(mfcdev->uar_map);
	mlx4_uar_free(dev, &mfcdev->priv_uar);
	mlx4_mr_free(dev, &mfcdev->mr);
	mlx4_pd_free(dev, mfcdev->priv_pdn);

	kfree(mfcdev);

	FCLOG(fc_log_ctrl, "%.*s: removed\n", MLX4_BOARD_ID_LEN, dev->board_id);
}

static void mlx4_fc_event(struct mlx4_dev *dev, void *mfcdev_ptr,
			  enum mlx4_dev_event event, int port)
{
	FCLOG(fc_log_ctrl, "%s %s: mlx4 event: %d\n",
	      dev_driver_string(&dev->pdev->dev), dev->pdev->dev.bus_id, event);
}

static inline struct mlx4_fc_vhba *find_vhba_for_netdev(struct net_device
							*netdev)
{
	struct mlx4_fc_dev *mfcdev;
	struct mlx4_fc_port *fc_port;
	struct mlx4_fc_vhba *vhba;
	int p;
	unsigned long flags2;

	spin_lock_irqsave(&mfcdev_list_lock, flags2);
	list_for_each_entry(mfcdev, &mfcdev_list, list)
	    for (p = 1; p <= MLX4_MAX_PORTS; ++p) {
		unsigned long flags;
		fc_port = &mfcdev->fc_port[p];
		if (! fc_port->initialized)
			continue;
		spin_lock_irqsave(&fc_port->vhba_list_lock, flags);
		list_for_each_entry(vhba, &fc_port->vhba_list, list)
		    if (vhba->underdev == netdev) {
			spin_unlock_irqrestore(&fc_port->vhba_list_lock, flags);
			spin_unlock_irqrestore(&mfcdev_list_lock, flags2);
			return vhba;
		}
		spin_unlock_irqrestore(&fc_port->vhba_list_lock, flags);
	}
	spin_unlock_irqrestore(&mfcdev_list_lock, flags2);
	return NULL;
}

extern spinlock_t cq_poll;

struct mlx4_dev *mlx4_en_find_netdev(struct net_device *netdev, int *port);

static inline struct net_device *vhba_get_netdev(struct mlx4_fc_vhba *vhba)
{
	return (struct net_device *)vhba->underdev;
}

static void mlx4_fc_link_change(struct mlx4_fc_vhba *vhba, int link_up)
{
	struct fc_lport *lp = vhba->lp;

	if (link_up)
		fc_linkup(lp);
	else
		fc_linkdown(lp);
}

static int mlx4_fc_register_netdev(struct net_device *netdev, int vlan_id,
				   int prio)
{
	struct mlx4_fc_vhba *vhba;
	struct mlx4_dev *mdev;
	int port;
	struct mlx4_fc_dev *mfcdev;
	int err;
	unsigned long flags;

	vhba = find_vhba_for_netdev(netdev);
	if (vhba) {
		mlx4_info(vhba->fport->mfcdev->dev,
			  "warning: already got vhba for %s. skipping\n",
			  netdev->name);
		return 0;
	}

	if (netdev->priv_flags & IFF_802_1Q_VLAN)
		mdev = mlx4_query_interface(vlan_dev_real_dev(netdev), &port);
	else
		mdev = mlx4_query_interface(netdev, &port);

	if (mdev) {
		FCLOG(fc_log_ctrl,
		      "%s belongs to mlx4_en. port=%d\n", netdev->name, port);
		spin_lock_irqsave(&mfcdev_list_lock, flags);
		list_for_each_entry(mfcdev, &mfcdev_list, list) {
			if (mfcdev->dev == mdev) {
				spin_unlock_irqrestore(&mfcdev_list_lock,
						       flags);
				err =
				    mlx4_fc_create_vhba(&mfcdev->fc_port[port],
							vlan_id, prio, netdev,
							netdev->name);
				if (err)
					mlx4_err(mdev,
						 "Could not create vhba"
						 " for net device %s vlan %d\n",
						 netdev->name, vlan_id);
				return 0;
			}
		}
		spin_unlock_irqrestore(&mfcdev_list_lock, flags);
		printk("No %s device in our device list\n", netdev->name);
	} else
		printk("%s does not belong to mlx4_en.\n", netdev->name);

	return -1;
}

static int mlx4_fc_unregister_netdev(struct net_device *netdev)
{
	struct mlx4_fc_vhba *vhba;
	unsigned long flags;

	vhba = find_vhba_for_netdev(netdev);
	if (!vhba) {
		FCLOG(fc_log_ctrl, "No vhba for %s. skipping.\n", netdev->name);
		return 0;
	}

	spin_lock_irqsave(&mfcdev_list_lock, flags);
	mlx4_fc_remove_vhba(vhba);
	spin_unlock_irqrestore(&mfcdev_list_lock, flags);

	return 0;
}

static int mlx4_fc_device_notification(struct notifier_block *notifier,
				       ulong event, void *ptr)
{
	struct net_device *netdev = ptr;
	struct mlx4_fc_vhba *vhba;

	vhba = find_vhba_for_netdev(netdev);

	switch (event) {
	case NETDEV_DOWN:
	case NETDEV_GOING_DOWN:
		if (!vhba)
			return NOTIFY_DONE;
		FCLOG(fc_log_ctrl, "notification: %s link down\n",
		      netdev->name);
		mlx4_fc_link_change(vhba, 0 /* link down */ );
		return NOTIFY_OK;
	case NETDEV_UP:
	case NETDEV_CHANGE:
		if (!vhba)
			return NOTIFY_DONE;
		FCLOG(fc_log_ctrl, "notification: %s link up\n", netdev->name);
		mlx4_fc_link_change(vhba, 1 /* link up */ );
		return NOTIFY_OK;
	case NETDEV_REGISTER:
		FCLOG(fc_log_ctrl, "registered netdev: %s\n", netdev->name);
		return NOTIFY_OK;
	case NETDEV_UNREGISTER:
		FCLOG(fc_log_ctrl, "unregister netdev: %s\n", netdev->name);
		return NOTIFY_OK;
	default:
		printk("unknown netdev event: %ld\n", event);
		return NOTIFY_OK;
	}
}

static struct notifier_block notifier_ops = {
	.notifier_call = mlx4_fc_device_notification,
};

static struct mlx4_interface mlx4_fc_interface = {
	.add = mlx4_fc_add,
	.remove = mlx4_fc_remove,
	.event = mlx4_fc_event
};

static void trimstr(char *str, int len)
{
	char *cp = str + len;
	while (--cp >= str && *cp == '\n')
		*cp = '\0';
}

static ssize_t mlx4_fc_sys_destroy(struct class *cl, const char *buf,
				   size_t count)
{
	char ifname[IFNAMSIZ];
	struct net_device *netdev = NULL;

	strncpy(ifname, buf, sizeof(ifname));
	trimstr(ifname, strlen(ifname));

	FCLOG(fc_log_ctrl, "Destoying vhba for interface %s\n", ifname);

	netdev = dev_get_by_name(&init_net, ifname);
	if (!netdev) {
		FCPRINT("Couldn't get a network device for '%s'", ifname);
		goto out;
	}

	mlx4_fc_unregister_netdev(netdev);

out:
	if (netdev)
		dev_put(netdev);
	return count;
}

static CLASS_ATTR(destroy, 0222, NULL, mlx4_fc_sys_destroy);

static ssize_t mlx4_fc_sys_create(struct class *cl, const char *buf,
				  size_t count)
{
	char ifname[IFNAMSIZ + 1];
	char *ch;
	char test;
	int cnt = 0;
	int vlan_id = -1;
	int prio = 0;
	struct net_device *netdev = NULL;

	strncpy(ifname, buf, sizeof(ifname));
	trimstr(ifname, strlen(ifname));

	if (ch = strchr(ifname, ',')) {
		*ch = '\0';
		cnt = sscanf(ch + 1, "%d%c", &prio, &test);
		if (cnt != 1 || prio < 0 || prio > 7) {
			printk("sysfs create: given priority is out of range 0-7"
			       ", assuming 0\n");
			prio = 0;
		}
	}

	FCLOG(fc_log_ctrl, "sysfs create: '%s'\n", ifname);

	netdev = dev_get_by_name(&init_net, ifname);
	if (!netdev) {
		FCPRINT("Couldn't get a network device for '%s'\n", ifname);
		goto out;
	}
	if (netdev->priv_flags & IFF_802_1Q_VLAN) {
		vlan_id = vlan_dev_vlan_id(netdev);
		FCLOG(fc_log_ctrl, "sysfs create: this is vlan id %d prio %d\n",
		      vlan_id, prio);
	}

	mlx4_fc_register_netdev(netdev, vlan_id, prio);

out:
	if (netdev)
		dev_put(netdev);
	return count;
}

static CLASS_ATTR(create, 0222, NULL, mlx4_fc_sys_create);

static ssize_t mlx4_fc_sys_link_change(struct class *cl, const char *buf,
				       size_t count, int link_up)
{
	struct mlx4_fc_vhba *vhba;
	struct net_device *netdev = NULL;
	char ifname[IFNAMSIZ];

	strncpy(ifname, buf, sizeof(ifname));
	trimstr(ifname, strlen(ifname));

	netdev = dev_get_by_name(&init_net, ifname);
	if (!netdev) {
		FCPRINT("Couldn't get a network device for '%s'", ifname);
		goto out;
	}

	vhba = find_vhba_for_netdev(netdev);
	if (!vhba) {
		FCPRINT("vhba for '%s' doesn't exist - ignoring\n", ifname);
		goto out;
	}

	mlx4_fc_link_change(vhba, link_up);

out:
	if (netdev)
		dev_put(netdev);
	return count;
}

static ssize_t mlx4_fc_sys_link_up(struct class *cl, const char *buf,
				   size_t count)
{
	return mlx4_fc_sys_link_change(cl, buf, count, 1);
}

static CLASS_ATTR(link_up, 0222, NULL, mlx4_fc_sys_link_up);

static ssize_t mlx4_fc_sys_link_down(struct class *cl, const char *buf,
				     size_t count)
{
	return mlx4_fc_sys_link_change(cl, buf, count, 0);
}

static CLASS_ATTR(link_down, 0222, NULL, mlx4_fc_sys_link_down);

struct class *mlx4_fc_class;

struct class_attribute *class_attrs[] = {
	&class_attr_link_up,
	&class_attr_link_down,
	&class_attr_create,
	&class_attr_destroy,
	NULL
};

#define MLX4_FC_MAJOR 271
static int __init mlx4_fc_init(void)
{
	int err;
	int i;

	FCLOG(fc_log_ctrl, "module loaded\n");

	if (mlx4_debug_mode == 1) {
		int r;
		r = sscanf(mlx4_gw_mac,
			   "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
			   &gw_mac[0], &gw_mac[1], &gw_mac[2], &gw_mac[3],
			   &gw_mac[4], &gw_mac[5]);
		if (r == 6) {
			FCLOG_MAC(fc_log_ctrl, "Setting GW MAC to: ", gw_mac);
		} else {
			printk("Bad gw_mac: %s. r=%d\n", mlx4_gw_mac, r);
			return -1;
		}
	}

	mlx4_fc_transport_template =
	    fc_attach_transport(&mlx4_fc_transport_function);

	if (mlx4_fc_transport_template == NULL) {
		printk("fail to attach fc transport");
		return -1;
	}

	/* mlx4_enable_wc(); */
	init_nport_id_table();
	err = mlx4_register_interface(&mlx4_fc_interface);
	if (err)
		return err;

	mlx4_fc_fcoe_wq = create_singlethread_workqueue("mlx4_fcoe");
	if (!mlx4_fc_fcoe_wq)
		goto err_unreg;

	err = register_netdevice_notifier(&notifier_ops);
	if (err)
		printk("could not register to netdevice notification chain\n");

	mlx4_fc_class = class_create(THIS_MODULE, "mlx4_fc");
	if (IS_ERR(mlx4_fc_class)) {
		goto err_unreg;
	}

	for (i = 0; class_attrs[i]; i++) {
		err = class_create_file(mlx4_fc_class, class_attrs[i]);
		if (err) {
			class_attrs[i] = NULL;
			return err;
		}
	}

	return 0;

err_unreg:
	mlx4_unregister_interface(&mlx4_fc_interface);

	return err;
}

static void __exit mlx4_fc_cleanup(void)
{
	int i;

	for (i = 0; class_attrs[i]; i++)
		class_remove_file(mlx4_fc_class, class_attrs[i]);

	class_destroy(mlx4_fc_class);
	unregister_netdevice_notifier(&notifier_ops);
	destroy_workqueue(mlx4_fc_fcoe_wq);
	mlx4_unregister_interface(&mlx4_fc_interface);
	fc_release_transport(mlx4_fc_transport_template);
	/* mlx4_disable_wc(); */
}

module_init(mlx4_fc_init);
module_exit(mlx4_fc_cleanup);
