/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/delay.h>
#include <linux/if_vlan.h>

#include <scsi/scsi.h>
#include <scsi/libfc.h>
#include <scsi/fc/fc_fcoe.h>

#include <linux/mlx4/device.h>
#include <linux/mlx4/cmd.h>

#include "mlx4_fc.h"

int mlx4_fc_frame_size = MLX4_DEFAULT_FC_MTU;
module_param_named(frame_size, mlx4_fc_frame_size, int, 0444);
MODULE_PARM_DESC(frame_size,
		 "Frame size default is " __stringify(MLX4_DEFAULT_FC_MTU));

int mlx4_fc_auto_start = 1;
module_param_named(auto_start, mlx4_fc_auto_start, int, 0444);
MODULE_PARM_DESC(auto_start, "Automatic FC login. Default = 1 (yes)");

int mlx4_fc_log_exch_per_vhba = MLX4_DEFAULT_LOG_EXCH_PER_VHBA;
module_param_named(log_exch_per_vhba, mlx4_fc_log_exch_per_vhba, int, 0444);
MODULE_PARM_DESC(log_exch_per_vhba,
		 "Max outstanding FC exchanges per virtual HBA (log). "
		 "Default =  " __stringify(MLX4_DEFAULT_LOG_EXCH_PER_VHBA));

int mlx4_fc_max_vhba_per_port = MLX4_DEFAULT_MAX_VHBA_PER_PORT;
module_param_named(max_vhba_per_port, mlx4_fc_max_vhba_per_port, int, 0444);
MODULE_PARM_DESC(max_vhba_per_port, "Max vHBAs allowed per port. "
		 "Default =  " __stringify(MLX4_DEFAULT_MAX_VHBA_PER_PORT));

struct libfc_function_template mlx4_libfc_fcn_templ = {
	.frame_send = mlx4_fc_send,
};

struct scsi_host_template mlx4_fc_driver_template = {
	.module = THIS_MODULE,
	.name = "MLX4_FCOE Driver",
	.proc_name = MLX4_FC_DRIVER_NAME,
	.queuecommand = mlx4_do_fcmd_send,
	.eh_host_reset_handler = fc_eh_host_reset,
	.slave_alloc = fc_slave_alloc,
	.change_queue_depth = fc_change_queue_depth,
	.this_id = -1,
	.cmd_per_lun = 32,
	.use_clustering = ENABLE_CLUSTERING,
	.sg_tablesize = MLX4_FC_MAX_FMR_PAGES,
	.max_sectors = MLX4_FC_MAX_FMR_PAGES,
/*	.eh_abort_handler = mlx4_fc_scsi_abort,
	.eh_device_reset_handler = mlx4_fc_device_reset,*/
};

static struct nport_id_entry
 fc_nport_id_table[MLX4_MAX_PORTS][MLX4_FC_NUM_NPORT_IDS];

void init_nport_id_table(void)
{
	memset(&fc_nport_id_table, 0, sizeof(fc_nport_id_table));
}

static u32 hw_index_to_key(u32 ind)
{
	return (ind >> 24) | (ind << 8);
}

static int mlx4_CONFIG_FC_BASIC(struct mlx4_dev *dev, u8 port,
				struct mlx4_fc_basic_config_params *params)
{
	struct mlx4_cmd_mailbox *mailbox;
	int err;

#define CONFIG_FC_FEXCH_BASE_OFFSET 	0x0
#define CONFIG_FC_NM_OFFSET		0x5
#define CONFIG_FC_NV_OFFSET		0x6
#define CONFIG_FC_NP_OOFSET		0x7
#define CONFIG_FC_BASEMPT_OFFSET	0x8
#define CONFIG_FC_NUM_RFCI_OFFSET	0xc
#define CONFIG_FC_RFCI_BASE_OFFSET	0xd
#define CONFIG_FC_PROMISC_QPN_OFFSET	0x14
#define CONFIG_FC_MCAST_QPN_OFFSET	0x18

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memset(mailbox->buf, 0, 256);

	MLX4_PUT(mailbox->buf, params->fexch_base, CONFIG_FC_FEXCH_BASE_OFFSET);
	MLX4_PUT(mailbox->buf, params->nm, CONFIG_FC_NM_OFFSET);
	MLX4_PUT(mailbox->buf, params->nv, CONFIG_FC_NV_OFFSET);
	MLX4_PUT(mailbox->buf, params->np, CONFIG_FC_NP_OOFSET);
	MLX4_PUT(mailbox->buf, (hw_index_to_key(params->fexch_base_mpt)),
		 CONFIG_FC_BASEMPT_OFFSET);
	MLX4_PUT(mailbox->buf,
		 params->rfci_base | (((u32) params->log_num_rfci) << 24),
		 CONFIG_FC_NUM_RFCI_OFFSET);
	MLX4_PUT(mailbox->buf, params->def_fcoe_promisc_qpn,
		 CONFIG_FC_PROMISC_QPN_OFFSET);
	MLX4_PUT(mailbox->buf, params->def_fcoe_mcast_qpn,
		 CONFIG_FC_MCAST_QPN_OFFSET);

	err = 0;
	err = mlx4_cmd(dev, mailbox->dma,
		       MLX4_CMD_INMOD_BASIC_CONF | port,
		       MLX4_CMD_MOD_FC_ENABLE,
		       MLX4_CMD_CONFIG_FC, MLX4_CMD_TIME_CLASS_B);

	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;

}

static int mlx4_CONFIG_FC_NPORT_ID(struct mlx4_dev *dev, u8 port,
				   struct nport_id_entry *table)
{
	struct mlx4_cmd_mailbox *mailbox;
	int err;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memcpy(mailbox->buf, table, MLX4_FC_NUM_NPORT_IDS * sizeof(u32));

	err = 0;

	err = mlx4_cmd(dev, mailbox->dma,
		       MLX4_CMD_INMOD_NPORT_TAB | port,
		       MLX4_CMD_MOD_FC_ENABLE,
		       MLX4_CMD_CONFIG_FC, MLX4_CMD_TIME_CLASS_B);

	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;
}

static int mlx4_fc_config_nport_id(struct mlx4_fc_port *fc_port, int idx,
				   u8 nport_id[3])
{
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	int port = fc_port->port;

	if ((idx < 0) || (idx >= MLX4_FC_NUM_NPORT_IDS))
		return -1;

	memcpy(fc_nport_id_table[port - 1][idx].fid, nport_id, 3);
	return mlx4_CONFIG_FC_NPORT_ID(mfcdev->dev, port,
				       fc_nport_id_table[port - 1]);
}

static int mlx4_QUERY_PORT(struct mlx4_dev *dev, u8 port, int *link_up)
{
	struct mlx4_cmd_mailbox *mailbox;
	struct mlx4_query_port_context *context;
	int err = 0;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	err = mlx4_cmd_box(dev, 0, mailbox->dma, port, 0, MLX4_CMD_QUERY_PORT,
			   MLX4_CMD_TIME_CLASS_B);
	if (err) {
		mlx4_err(dev, "Query port returned error: %d\n", err);
		goto query_err;
	}

	context = (struct mlx4_query_port_context *)mailbox->buf;

	/* this is a double-checking... mlx4_core has done it for us */
	if (!(context->supported_port_type & MLX4_PORT_TYPE_ETH)) {
		mlx4_err(dev, "Port does not support ETH\n");
		goto query_err;
	}

	*link_up = context->supported_port_type & QUERY_PORT_LINK_MASK;

query_err:
	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;
}

int mlx4_fc_flogi_finished(struct fc_lport *lp)
{
	struct mlx4_fc_vhba *vhba = lport_priv(lp);
	struct mlx4_fc_port *fc_port = vhba->fport;
	int err = 0;
	u8 fid[3];

	memcpy(fid, vhba->my_fid, 3);

	FCLOG(fc_log_ctrl,
	      "port%d vhba%d: Finishing flogi process \n", fc_port->port,
	      vhba->idx);
	if (!mlx4_debug_mode) {
		FCLOG(fc_log_ctrl,
		      "port%d vhba%d: Starting DATA RFCI\n",
		      fc_port->port, vhba->idx);
		err = mlx4_start_rfci_data(vhba,
					   mac_to_u64(vhba->rfci[RFCI_DATA].
						      mac));
		if (err) {
			mlx4_err(fc_port->mfcdev->dev,
				 "port%d vhba%d: Could not start DATA RFCI, err=%d\n",
				 fc_port->port, vhba->idx, err);
			goto err;
		}
		FCLOG_MAC(fc_log_ctrl,
			  "port%d vhba%d: DATA RFCI started with MAC ",
			  vhba->rfci[RFCI_DATA].mac, fc_port->port, vhba->idx);

	}
	err = mlx4_fc_init_fcmd(vhba);
	if (err) {
		mlx4_err(fc_port->mfcdev->dev,
			 "port%d vhba%d: Could not init FCMD, err=%d\n",
			 fc_port->port, vhba->idx, err);
		goto err;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: FCMD in state RTS\n", fc_port->port,
	      vhba->idx);

	err = mlx4_fc_config_nport_id(fc_port, 0, fid);
	if (err) {
		mlx4_err(fc_port->mfcdev->dev,
			 "port%d vhba%d: Could not config nport id %x:%x:%x\n",
			 fc_port->port, vhba->idx, fid[0], fid[1], fid[2]);
		goto err;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: NPort ID %x:%x:%x\n", fc_port->port,
	      vhba->idx, fid[0], fid[1], fid[2]);

err:
	return err;
}

static int libfc_config(struct fc_lport *lp)
{
	lp->tt = mlx4_libfc_fcn_templ;

	if (fc_fcp_init(lp))
		return -ENOMEM;
	fc_exch_init(lp);
	fc_elsct_init(lp);
	fc_lport_init(lp);
	fc_rport_init(lp);
	fc_disc_init(lp);

	return 0;
}

static void mlx4_fc_lport_destroy(struct fc_lport *lp)
{
	int cpu;

	for_each_online_cpu(cpu) {
		kfree(lp->dev_stats[cpu]);
		lp->dev_stats[cpu] = NULL;
	}

	fc_exch_mgr_free(lp->emp);
	lp->emp = NULL;
}

static int mlx4_fc_lport_config(struct fc_lport *lp)
{
	int i = 0;

	lp->link_up = 0;
	lp->qfull = 0;
	lp->max_retry_count = 3;
	lp->e_d_tov = 2 * 1000;
	lp->r_a_tov = 2 * 2 * 1000;
	lp->max_retry_count = 3;
	lp->service_params = (FCP_SPPF_INIT_FCN | FCP_SPPF_RD_XRDY_DIS |
			      FCP_SPPF_RETRY | FCP_SPPF_CONF_COMPL);
	lp->link_supported_speeds |= FC_PORTSPEED_1GBIT | FC_PORTSPEED_4GBIT |
	    FC_PORTSPEED_10GBIT | FC_PORTSPEED_8GBIT | FC_PORTSPEED_16GBIT;
	lp->link_speed = FC_PORTSPEED_10GBIT;

	for_each_online_cpu(i) {
		lp->dev_stats[i] =
		    kzalloc(sizeof(struct fcoe_dev_stats), GFP_KERNEL);
		if (!lp->dev_stats[i])
			goto err_out;
	}

	fc_lport_config(lp);

	lp->emp = fc_exch_mgr_alloc(lp, FC_CLASS_3,
				    MLX4_FC_MIN_XID, MLX4_FC_MAX_XID);
	if (!lp->emp)
		return -ENOMEM;

	return 0;

err_out:
	for_each_online_cpu(i) {
		if (!lp->dev_stats[i])
			break;
		kfree(lp->dev_stats[i]);
		lp->dev_stats[i] = NULL;
	}

	lp->emp = NULL;
	return -ENOMEM;

}

static void mlx4_libfc_destroy(struct fc_lport *lp)
{
	fc_remove_host(lp->host);
	scsi_remove_host(lp->host);
	fc_lport_destroy(lp);
	fc_fcp_destroy(lp);
}

static int mlx4_libfc_init(struct fc_lport *lp, int auto_start,
			   int port, const char *symbolic_name)
{
	struct mlx4_fc_vhba *vhba = lport_priv(lp);
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	int err;

	fc_set_wwnn(lp,
		    mfcdev->dev->caps.def_mac[fc_port->
					      port] | ((u64) 0x1000 << 48));
	fc_set_wwpn(lp,
		    mfcdev->dev->caps.def_mac[fc_port->
					      port] | ((u64) 0x2000 << 48));
	fc_set_mfs(lp, vhba->fc_frame_size);

	lp->host->max_lun = MLX4_FC_MAX_LUN;
	lp->host->max_id = MLX4_FC_MAX_FCP_TARGET;
	lp->host->max_channel = 0;
	lp->host->transportt = mlx4_fc_transport_template;

	err = scsi_add_host(lp->host, NULL);
	if (err) {
		FCPRINT("error on scsi_add_host\n");
		return err;
	}

	snprintf(fc_host_symbolic_name(lp->host), FC_SYMBOLIC_NAME_SIZE,
		 "%s v%s over %s",
		 MLX4FC_DRIVER_NAME, MLX4FC_DRIVER_VERSION, symbolic_name);

	err = libfc_config(lp);
	if (err) {
		FCPRINT("error on libfc_config\n");
		return err;
	}

	fc_fabric_login(lp);

	return 0;
}

/* like fcoe_create_interface */
int mlx4_fc_create_vhba(struct mlx4_fc_port *fc_port, int vlan_id, int prio,
			void *underdev, const char *symbolic_name)
{
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_caps *caps = &mfcdev->dev->caps;
	struct fc_lport *lp;
	struct mlx4_fc_vhba *vhba;
	int idx, port = fc_port->port;
	int err;
	unsigned long flags;
	struct Scsi_Host *shost;

	FCLOG(fc_log_ctrl, "port%d: Creating new vhba for vlan_id=%d\n",
	      port, vlan_id);

	mlx4_fc_driver_template.can_queue = 1 << mlx4_fc_log_exch_per_vhba;

	shost = scsi_host_alloc(&mlx4_fc_driver_template,
				sizeof(struct fc_lport) +
				sizeof(struct mlx4_fc_vhba));
	if (!shost) {
		FCPRINT("Could not allocate host structure\n");
		err = -ENOMEM;
		goto err_out;
	}

	lp = shost_priv(shost);
	lp->host = shost;
	vhba = lport_priv(lp);
	vhba->lp = lp;

	err = mlx4_fc_lport_config(lp);
	if (err) {
		FCPRINT("Error configuring lport\n");
		goto err_host_put;
	}

	idx = mlx4_fc_bitmap_slot_alloc(&fc_port->fexch_bulk_bm);
	if (idx == -1) {
		mlx4_err(mfcdev->dev,
			 "port%d: fail alloc fexch bulk for new vhba\n", port);
		err = -ENOMEM;
		goto err_lport_destroy;
	}
	vhba->idx = idx;
	vhba->fport = fc_port;
	vhba->underdev = underdev;
	vhba->rfci[RFCI_DATA].fc_mac_idx = -1;
	/* TODO: needed? */
	vhba->rfci_rx_enabled = 0;
	/* sizeof(struct fcoe_hdr) */
	vhba->fcoe_hlen = sizeof(struct fcoe_hdr_old);
	vhba->num_fexch = 1 << fc_port->log_num_fexch_per_vhba;
	vhba->base_fexch_qpn = fc_port->base_fexch_qpn + idx * vhba->num_fexch;
	vhba->base_fexch_mpt = fc_port->base_fexch_mpt + idx * vhba->num_fexch;

	FCLOG(fc_log_ctrl,
	      "port%d vhba%d: b_fexch_qpn=0x%x, b_fexch_mpt=0x%x, n_fexch=%d\n",
	      port, idx, vhba->base_fexch_qpn, vhba->base_fexch_mpt,
	      vhba->num_fexch);

	/* Configure port */
	vhba->fc_frame_size = MIN(mlx4_fc_frame_size,
				  (caps->eth_mtu_cap[port] -
				   14 /*mac */  - 2 /*fcoe */  - 24 /*fc */  -
				   5 /*crceof */ ) & ~3);

	fc_set_mfs(lp, vhba->fc_frame_size);

	FCLOG(fc_log_ctrl,
	      "port%d vhba%d: Setting port FC frame size to %d\n",
	      port, idx, vhba->fc_frame_size);

	vhba->fc_vlan_id = vlan_id;
	vhba->fc_vlan_prio = prio;
	if (vlan_id != -1) {
		err = mlx4_register_vlan(mfcdev->dev, port, vlan_id,
					 &vhba->fc_vlan_idx);
		if (err) {
			mlx4_err(mfcdev->dev,
				 "port%d vhba%d: fail to reg VLAN %d. err=%x\n",
				 port, idx, vlan_id, err);
			goto err_free_fexch_bulk;
		}
		FCLOG(fc_log_ctrl,
		      "port%d vhba%d: Reg vlan %d prio %d to index %d\n",
		      port, idx, vlan_id, prio, vhba->fc_vlan_idx);
	}

	u64_to_mac(vhba->rfci[RFCI_CTRL].mac, caps->def_mac[port]);

	err =
	    mlx4_fc_create_rfci(vhba, &vhba->rfci[RFCI_CTRL],
				caps->def_mac[port]);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not create CTRL RFCI, err=%d\n",
			 port, idx, err);
		goto err_unreg_vlan;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: Created CTRL RFCI\n", port, idx);

	err = mlx4_fc_create_rfci(vhba, &vhba->rfci[RFCI_FLOGI],
				  mac_to_u64((u8[6]) FC_FCOE_FLOGI_MAC));
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: fail to create FLOGI RFCI, err=%d\n",
			 port, idx, err);
		goto err_destroy_rfci_ctrl;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: Created FLOGI RFCI\n", port, idx);

	err = mlx4_fc_create_fcmd(vhba);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not create FCMD, err=%d\n",
			 port, idx, err);
		goto err_destroy_rfci_flogi;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: FCMD created\n", port, idx);

	err = mlx4_libfc_init(lp, mlx4_fc_auto_start, port, symbolic_name);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not init libfc device\n",
			 port, idx);
		goto err_destroy_fcmd;
	}

	err = mlx4_fc_init_rfci(vhba, &vhba->rfci[RFCI_CTRL]);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not init CTRL RFCI, err=%d\n",
			 port, idx, err);
		goto err_destroy_libfc;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: CTRL RFCI in state RTS\n", port,
	      idx);

	err = mlx4_fc_init_rfci(vhba, &vhba->rfci[RFCI_FLOGI]);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d vhba%d: Could not init FLOGI RFCI, err=%d\n",
			 port, idx, err);
		goto err_destroy_libfc;
	}
	FCLOG(fc_log_ctrl, "port%d vhba%d: FLOGI RFCI in state RTS\n", port,
	      idx);

	memcpy(vhba->dest_addr, gw_mac, ETH_ALEN);

	spin_lock_irqsave(&fc_port->vhba_list_lock, flags);
	list_add(&vhba->list, &fc_port->vhba_list);
	spin_unlock_irqrestore(&fc_port->vhba_list_lock, flags);

	FCLOG(fc_log_ctrl, "port%d vhba%d: ready\n", port, idx);

	return 0;

err_destroy_libfc:
	mlx4_libfc_destroy(lp);
err_destroy_fcmd:
	mlx4_fc_destroy_fcmd(vhba);
err_destroy_rfci_flogi:
	mlx4_fc_destroy_rfci(vhba, &vhba->rfci[RFCI_FLOGI]);
err_destroy_rfci_ctrl:
	mlx4_fc_destroy_rfci(vhba, &vhba->rfci[RFCI_CTRL]);
err_unreg_vlan:
	if (vhba->fc_vlan_idx != -1)
		mlx4_unregister_vlan(mfcdev->dev, port, vhba->fc_vlan_idx);
err_free_fexch_bulk:
	mlx4_fc_bitmap_slot_free(&fc_port->fexch_bulk_bm, idx);
err_lport_destroy:
	mlx4_fc_lport_destroy(lp);
err_host_put:
	scsi_host_put(lp->host);
err_out:
	return err;
}

/* vhba->fport->vhba_list_lock must be held */
void mlx4_fc_remove_vhba(struct mlx4_fc_vhba *vhba)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	int port = fc_port->port, idx = vhba->idx;
	struct fc_lport *lp = vhba->lp;

	FCLOG(fc_log_ctrl, "port%d vhba%d: removing vhba\n", port, idx);

	list_del(&vhba->list);

	/* fc_linkdown(vhba->lp); */
	mlx4_libfc_destroy(vhba->lp);
	mlx4_fc_destroy_fcmd(vhba);
	mlx4_fc_destroy_rfci(vhba, &vhba->rfci[RFCI_FLOGI]);
	mlx4_fc_destroy_rfci(vhba, &vhba->rfci[RFCI_CTRL]);
	if (vhba->rfci[RFCI_DATA].fc_mac_idx != -1)
		mlx4_fc_destroy_rfci(vhba, &vhba->rfci[RFCI_DATA]);
	if (vhba->fc_vlan_idx != -1)
		mlx4_unregister_vlan(mfcdev->dev, port, vhba->fc_vlan_idx);
	mlx4_fc_bitmap_slot_free(&fc_port->fexch_bulk_bm, idx);

	mlx4_fc_lport_destroy(lp);
	scsi_host_put(lp->host);
	FCLOG(fc_log_ctrl, "port%d vhba%d: removed vhba\n", port, idx);
}

int mlx4_fc_init_port(struct mlx4_fc_dev *mfcdev, int port)
{
	struct mlx4_fc_port *fport = &mfcdev->fc_port[port];
	int err = 0;
	int mvp = (1 << mfcdev->log_num_mac) * (1 << mfcdev->log_num_vlan) *
	    (1 << mfcdev->log_num_prio);
	int qpn_offset;
	struct mlx4_fc_basic_config_params params = { 0 };

	FCLOG(fc_log_ctrl, "port%d: is ETH, setting up for FCoE.\n", port);
	if (mfcdev->log_num_mac < 2) {
		mlx4_err(mfcdev->dev, "Not enough MAC addresses allowed in "
			 "mlx4_core. Need log_num_mac >= 2\n");
		return -EINVAL;
	}

	fport->port = port;
	fport->mfcdev = mfcdev;
	fport->vhba_list_lock = __SPIN_LOCK_UNLOCKED(fport->vhba_list_lock);
	INIT_LIST_HEAD(&fport->vhba_list);
	fport->num_fexch_qps = (1 << mlx4_fc_log_exch_per_vhba) *
	    mlx4_fc_max_vhba_per_port;
	fport->log_num_fexch_per_vhba = mlx4_fc_log_exch_per_vhba;
	err = mlx4_qp_reserve_range(mfcdev->dev, fport->num_fexch_qps,
				    MLX4_FC_MAX_PORT_FEXCH,
				    &fport->base_fexch_qpn);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate QP range for FEXCH."
			 " Need 0x%x QPs aligned to 0x%x\n",
			 fport->num_fexch_qps, MLX4_FC_MAX_PORT_FEXCH);
		err = -ENOMEM;
		goto err_out;
	}
	/* TODO: for bidirectional SCSI we'll need to double the amount of
	   reserved MPTs, with proper spanning */
	err = mlx4_mr_reserve_range(mfcdev->dev, fport->num_fexch_qps,
				    2 * MLX4_FC_MAX_PORT_FEXCH,
				    &fport->base_fexch_mpt);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate MPT range for FEXCH."
			 " Need 0x%x MPTs aligned to 0x%x\n",
			 fport->num_fexch_qps, 2 * MLX4_FC_MAX_PORT_FEXCH);
		err = -ENOMEM;
		goto err_free_qp_range;
	}

	qpn_offset = (port - 1) * fport->num_fexch_qps;

	fport->base_rfci_qpn = mfcdev->base_rfci_qpn + (port - 1) * mvp;
	params.rfci_base = fport->base_rfci_qpn;
	params.fexch_base = fport->base_fexch_qpn;
	params.fexch_base_mpt = fport->base_fexch_mpt;
	params.nm = fport->n_m = mfcdev->log_num_mac;
	params.nv = fport->n_v = mfcdev->log_num_vlan;
	params.np = fport->n_p = mfcdev->log_num_prio;
	params.log_num_rfci = ilog2(mvp);
	params.def_fcoe_promisc_qpn = 0x77;
	params.def_fcoe_mcast_qpn = 0x78;

	FCLOG(fc_log_ctrl,
	      "port%d: b_fexch=0x%x, n_fexch=0x%x, b_mpt=0x%x,"
	      " b_rfci=0x%x, num_rfci=0x%x\n",
	      port, fport->base_fexch_qpn, fport->num_fexch_qps,
	      fport->base_fexch_mpt, fport->base_rfci_qpn, mvp);

	err = mlx4_CONFIG_FC_BASIC(mfcdev->dev, port, &params);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d: Could not issue CONFIG_FC Basic command\n",
			 port);
		goto err_free_mr_range;
	}
	FCLOG(fc_log_ctrl, "port%d: CONFIG_FC Basic successfull\n", port);

	err = mlx4_fc_bitmap_alloc(&fport->fexch_bulk_bm,
				   fport->num_fexch_qps >> fport->
				   log_num_fexch_per_vhba);
	if (err) {
		mlx4_err(mfcdev->dev,
			 "port%d: Could not allocate fexch bulks bitmap\n",
			 port);
		goto err_free_mr_range;
	}

	fport->initialized = 1;

	return 0;

err_free_qp_range:
	mlx4_qp_release_range(mfcdev->dev, fport->base_fexch_qpn,
			      fport->num_fexch_qps);
err_free_mr_range:
	mlx4_mr_release_range(mfcdev->dev, fport->base_fexch_mpt,
			      fport->num_fexch_qps);
err_out:
	return err;
}

void mlx4_fc_free_port(struct mlx4_fc_dev *mfcdev, int port)
{
	struct mlx4_fc_port *fc_port = &mfcdev->fc_port[port];
	struct mlx4_fc_vhba *vhba, *tmp;
	unsigned long flags;

	FCLOG(fc_log_ctrl, "port%d: freeing\n", port);

	spin_lock_irqsave(&fc_port->vhba_list_lock, flags);
	list_for_each_entry_safe(vhba, tmp, &fc_port->vhba_list, list)
	    mlx4_fc_remove_vhba(vhba);
	spin_unlock_irqrestore(&fc_port->vhba_list_lock, flags);

	/* 
	 * make sure the bitmap is empty, meaning, no vhba's left using
	 * fexch bulk
	 */
	ASSERT(mlx4_fc_bitmap_empty(&fc_port->fexch_bulk_bm));
	mlx4_fc_bitmap_free(&fc_port->fexch_bulk_bm);
	mlx4_qp_release_range(mfcdev->dev, fc_port->base_fexch_qpn,
			      fc_port->num_fexch_qps);
	mlx4_mr_release_range(mfcdev->dev, fc_port->base_fexch_mpt,
			      fc_port->num_fexch_qps);
	fc_port->initialized = 0;
	FCLOG(fc_log_ctrl, "port%d: free\n", port);
}
