/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/etherdevice.h>

#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/qp.h>
#include <linux/mlx4/cq.h>

#include <scsi/libfc.h>

#include "mlx4_fc.h"

#define MLX4_CQE_QPN_MASK 0x00ffffff

static void mlx4_fc_rx_rfci(struct work_struct *work);

static int mlx4_fc_prepare_rx_buf(struct mlx4_fc_vhba *vhba,
				  struct mlx4_fc_rfci *rfci)
{
	struct mlx4_fc_q *rq = &rfci->fc_qp.rq;
	struct sk_buff *skb;
	struct mlx4_fc_rfci_rx_info *fr;
	int index;

	skb = dev_alloc_skb(MLX4_FC_RFCI_RX_SKB_BUFSIZE +
			    sizeof(struct mlx4_fc_rfci_rx_info));
	if (!skb) {
		FCPRINT("No mem - rx packet dropped\n");
		goto err_out;
	}

	skb_reserve(skb, sizeof(struct mlx4_fc_rfci_rx_info));

	fr = mlx4_fc_rfci_rx_info_from_skb(skb);
	fr->vhba = vhba;
	fr->skb = skb;

	index = mlx4_fc_post_rx_buf(vhba->fport->mfcdev, &rfci->fc_qp,
				    skb->data, MLX4_FC_RFCI_RX_SKB_BUFSIZE);

	mlx4_fc_q_info_get(rq, index, struct sk_buff *) = skb;

	return 0;

err_out:
	return -1;
}

static void mlx4_fc_rfci_unpost_rx_bufs(struct mlx4_fc_dev *mfcdev,
					struct mlx4_fc_q *rq)
{
	int i;
	unsigned long flags;

	spin_lock_irqsave(&rq->lock, flags);
	for (i = 0; i < rq->size; i++) {
		struct sk_buff *skb;
//              struct mlx4_fc_rx_desc *rx_desc = rq->buf + (i * rq->stride);

		skb = mlx4_fc_q_info_get(rq, i, struct sk_buff *);
		if (!skb)
			continue;
/* this causes oops - why? */
//              pci_unmap_single(mfcdev->dev->pdev,
//                               be64_to_cpu(rx_desc->data[0].addr),
//                               be32_to_cpu(rx_desc->data[0].count),
//                               PCI_DMA_FROMDEVICE);

		mlx4_fc_q_info_get(rq, i, struct sk_buff *) = NULL;

		kfree_skb(skb);
	}
	spin_unlock_irqrestore(&rq->lock, flags);
}

static struct mlx4_fc_rfci *rfci_by_qpn(struct mlx4_fc_vhba *vhba, int qpn)
{
	int i;

	for (i = 0; (i < RFCI_NUM) && (vhba->rfci[i].fc_qp.mqp.qpn != qpn);
	     i++) ;

	ASSERT_PR(i < RFCI_NUM, "Couldn't find matching rfci for qpn %d\n",
		  qpn);

	return &vhba->rfci[i];
}

static void mlx4_fc_rfci_tx_comp(struct mlx4_fc_vhba *vhba,
				 struct mlx4_cqe *cqe)
{
	struct mlx4_fc_rfci *rfci;
	struct mlx4_fc_q *sq;
	struct sk_buff *skb;
	struct mlx4_fc_rfci_tx_desc *tx_desc;
	u32 index;
	unsigned long flags;

	rfci = rfci_by_qpn(vhba,
			   be32_to_cpu(cqe->vlan_my_qpn) & MLX4_CQE_QPN_MASK);

	sq = &rfci->fc_qp.sq;
	index = be16_to_cpu(cqe->wqe_index) & sq->size_mask;
	tx_desc = sq->buf + index * RFCI_SQ_BB_SIZE;

	pci_unmap_single(vhba->fport->mfcdev->dev->pdev,
			 be64_to_cpu(tx_desc->data.addr),
			 be32_to_cpu(tx_desc->data.count), PCI_DMA_TODEVICE);

	skb = mlx4_fc_q_info_get(sq, index, struct sk_buff *);
	mlx4_fc_q_info_get(sq, index, struct sk_buff *) = NULL;

	spin_lock_irqsave(&sq->lock, flags);
	++sq->cons;
	spin_unlock_irqrestore(&sq->lock, flags);
}

static void mlx4_fc_rfci_rx_comp(struct mlx4_fc_vhba *vhba,
				 struct mlx4_cqe *cqe)
{
	struct mlx4_fc_rfci *rfci;
	struct mlx4_fc_q *rq;
	struct mlx4_fc_rx_desc *rx_desc;
	u32 index;
	int len;
	unsigned long flags;
	struct sk_buff *skb;
	struct mlx4_fc_rfci_rx_info *fr;
	int err;

	rfci = rfci_by_qpn(vhba,
			   be32_to_cpu(cqe->vlan_my_qpn) & MLX4_CQE_QPN_MASK);
	rq = &rfci->fc_qp.rq;
	index = be16_to_cpu(cqe->wqe_index) & rq->size_mask;
	rx_desc = rq->buf + (index * rq->stride);

	pci_unmap_single(vhba->fport->mfcdev->dev->pdev,
			 be64_to_cpu(rx_desc->data[0].addr),
			 be32_to_cpu(rx_desc->data[0].count),
			 PCI_DMA_FROMDEVICE);

	skb = mlx4_fc_q_info_get(rq, index, struct sk_buff *);
	mlx4_fc_q_info_get(rq, index, struct sk_buff *) = NULL;

	spin_lock_irqsave(&rq->lock, flags);
	rfci->fc_qp.rq.cons++;
	spin_unlock_irqrestore(&rq->lock, flags);

	len = be32_to_cpu(cqe->byte_cnt);

	fr = mlx4_fc_rfci_rx_info_from_skb(skb);
	skb_put(skb, len);
	skb_set_mac_header(skb, 0);
	skb_pull(skb, ETH_HLEN);

	if (!vhba->rfci_rx_enabled) {
		FCPRINT("RFCI RX ignored till host started\n");
		goto err_out;
	}

	INIT_WORK(&fr->work, mlx4_fc_rx_rfci);
	queue_work(mlx4_fc_fcoe_wq, &fr->work);

	err = mlx4_fc_prepare_rx_buf(vhba, rfci);
	if (err) {
		FCPRINT("No mem - rx packet dropped\n");
		goto err_out;
	}

	return;

      err_out:
	ASSERT_PR(0, "memory alloc failed - TODO: reuse skb\n");

	kfree_skb(skb);
}

int mlx4_fc_create_rfci(struct mlx4_fc_vhba *vhba, struct mlx4_fc_rfci *rfci,
			u64 mac)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_qp *qp = &rfci->fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	struct mlx4_fc_q *rq = &qp->rq;
	u32 qpn;
	int err = 0;
	int i;

	FCLOG(fc_log_ctrl, "creating RFCI for mac 0x%llx\n", mac);

	err = mlx4_register_mac(mfcdev->dev, fc_port->port, mac,
				&rfci->fc_mac_idx);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not register MAC\n");
		goto err_out;
	}

	err = mlx4_fc_q_init(sq, RFCI_SQ_BB_SIZE, RFCI_SQ_NUM_BBS,
			     sizeof(struct sk_buff *));
	if (err) {
		mlx4_err(mfcdev->dev, "Error initializing rfci sq\n");
		goto err_unreg_mac;
	}

	err = mlx4_fc_q_init(rq, RFCI_RQ_WQE_SIZE, RFCI_RQ_NUM_WQES,
			     sizeof(struct sk_buff *));
	if (err) {
		mlx4_err(mfcdev->dev, "Error initializing rfci rq\n");
		err = -ENOMEM;
		goto err_free_txinfo;
	}

	qp->buf_size = (sq->size * sq->stride) + (rq->size * rq->stride);

	err = mlx4_alloc_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size,
				 qp->buf_size);
	if (err)
		goto err_free_rxinfo;

	if (RFCI_SQ_BB_SIZE >= RFCI_RQ_WQE_SIZE) {
		sq->buf = qp->wqres.buf.direct.buf;
		rq->buf = sq->buf + (sq->size * sq->stride);
	} else {
		rq->buf = qp->wqres.buf.direct.buf;
		sq->buf = rq->buf + (rq->size * rq->stride);
	}

	*qp->wqres.db.db = 0;

	mlx4_fc_stamp_sq(sq);

	qpn = fc_port->base_rfci_qpn | 
	      (rfci->fc_mac_idx << (fc_port->n_v + fc_port->n_p));
						     
	if (vhba->fc_vlan_idx != -1)
		qpn |= (vhba->fc_vlan_idx << fc_port->n_p);

	FCLOG(fc_log_ctrl,
	      "using QPN=0x%x (mfcdev base is 0x%x, port base is 0x%x)\n", qpn,
	      mfcdev->base_rfci_qpn, fc_port->base_rfci_qpn);

	err = mlx4_qp_alloc(mfcdev->dev, qpn, &rfci->fc_qp.mqp);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate QP number 0x%x\n",
			 qpn);
		goto err_free_man;
	}

	qp->doorbell_qpn = swab32(qp->mqp.qpn << 8);

	err = mlx4_fc_create_cq(vhba, &rfci->fc_cq, MLX4_FC_RFCI_CQ_ENTRIES, 0, 1,	/* arm cq */
				mlx4_fc_rfci_rx_comp, mlx4_fc_rfci_tx_comp,
				"RFCI");
	if (err) {
		mlx4_err(mfcdev->dev,
			 "Failed creating RFCI CQ for port %d, err=%d\n",
			 fc_port->port, err);
		goto err_free_qp;
	}

	FCLOG(fc_log_ctrl,
	      "Created RFCI CQ successfull for port %d, CQN=0x%x\n",
	      fc_port->port, rfci->fc_cq.mcq.cqn);

	for (i = 0; i < rq->size - 1; i++) {
		err = mlx4_fc_prepare_rx_buf(vhba, rfci);
		if (err) {
			FCPRINT("Failed preparing RX descriptor[%d] for RFCI\n",
				i);
			goto err_free_cq;
		}
	}
	mlx4_fc_ring_db_rx(&rfci->fc_qp);

	return 0;

err_free_cq:
	mlx4_fc_rfci_unpost_rx_bufs(mfcdev, &rfci->fc_qp.rq);
	mlx4_fc_destroy_cq(&rfci->fc_cq);
err_free_qp:
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
err_free_man:
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
err_free_rxinfo:
	mlx4_fc_q_destroy(rq);
err_free_txinfo:
	mlx4_fc_q_destroy(sq);
err_unreg_mac:
	mlx4_unregister_mac(mfcdev->dev, fc_port->port, rfci->fc_mac_idx);
err_out:
	return err;
}

void mlx4_fc_destroy_rfci(struct mlx4_fc_vhba *vhba, struct mlx4_fc_rfci *rfci)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_qp *qp = &rfci->fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	struct mlx4_fc_q *rq = &qp->rq;

	mlx4_fc_destroy_cq(&rfci->fc_cq);
	if (qp->is_created)
		mlx4_qp_to_reset(mfcdev->dev, &qp->mqp);
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);

	mlx4_fc_rfci_unpost_rx_bufs(mfcdev, &rfci->fc_qp.rq);

	mlx4_fc_q_destroy(rq);
	mlx4_fc_q_destroy(sq);

	mlx4_unregister_mac(mfcdev->dev, fc_port->port, rfci->fc_mac_idx);
	rfci->fc_mac_idx = -1;
}

int mlx4_fc_init_rfci(struct mlx4_fc_vhba *vhba, struct mlx4_fc_rfci *rfci)
{
	struct mlx4_fc_port *fc_port = vhba->fport;
	struct mlx4_fc_dev *mfcdev = fc_port->mfcdev;
	struct mlx4_fc_qp *qp = &rfci->fc_qp;
	enum mlx4_qp_state qp_state = MLX4_QP_STATE_RST;
	int err = 0;

	struct mlx4_qp_context context = {
		.flags = cpu_to_be32(QPC_SERVICE_TYPE_RFCI << 16),
		.pd = cpu_to_be32(mfcdev->priv_pdn),
		/* Raw-ETH requirement */
		.mtu_msgmax = 0x77,
		.sq_size_stride =
		    ilog2(RFCI_SQ_NUM_BBS) << 3 | ilog2(RFCI_SQ_BB_SIZE >> 4),
		.rq_size_stride =
		    ilog2(RFCI_RQ_NUM_WQES) << 3 | ilog2(RFCI_RQ_WQE_SIZE >> 4),
		.usr_page = cpu_to_be32(mfcdev->priv_uar.index),
		.local_qpn = cpu_to_be32(qp->mqp.qpn),
		.pri_path.sched_queue =
		    0x83 | vhba->fc_vlan_prio << 3 | (fc_port->port - 1) << 6,
		.pri_path.counter_index = 0xff,
		.pri_path.ackto = MLX4_LINK_TYPE_ETH,
		.params2 =
		    cpu_to_be32((qp->wqres.buf.direct.
				 map & (PAGE_SIZE - 1)) & 0xfc0),
		.cqn_send = cpu_to_be32(rfci->fc_cq.mcq.cqn),
		.cqn_recv = cpu_to_be32(rfci->fc_cq.mcq.cqn),
		/* we can assume that db.dma is aligned */
		.db_rec_addr = cpu_to_be64(qp->wqres.db.dma),
		.srqn = 0,
	};

	err =
	    mlx4_qp_to_ready(mfcdev->dev, &qp->wqres.mtt, &context, &qp->mqp,
			     &qp_state);

	if (qp_state != MLX4_QP_STATE_RST)
		qp->is_created = 1;

	if (qp_state != MLX4_QP_STATE_RTS) {
		mlx4_err(mfcdev->dev, "Error bringing RFCI QP to RTS state\n");
		return err;
	}
	return 0;
}

int mlx4_do_rfci_xmit(struct mlx4_fc_vhba *vhba, int channel,
		      struct sk_buff *skb, u8 fceof)
{
	struct mlx4_fc_rfci *rfci = &vhba->rfci[channel];
	struct mlx4_fc_dev *mfcdev = vhba->fport->mfcdev;
	struct mlx4_dev *mdev = mfcdev->dev;
	struct mlx4_fc_q *sq = &rfci->fc_qp.sq;

	struct mlx4_fc_rfci_tx_desc *tx_desc;
	struct mlx4_fc_data_seg *data;

	unsigned int desc_size;
	unsigned int pad_size;
	dma_addr_t dma;
	u32 index, prod;
	__be32 op_own;
	unsigned long flags;
	desc_size = CTRL_SIZE + DS_SIZE;
	pad_size = RFCI_SQ_BB_SIZE - desc_size;

	spin_lock_irqsave(&sq->lock, flags);
	if (unlikely((u32) (sq->prod - sq->cons - 1) > sq->size - 2)) {
		mlx4_err(mdev, "rfci_xmit: Out of send queue BBs\n");
		spin_unlock_irqrestore(&sq->lock, flags);
		return 0;
	}
	prod = sq->prod;
	++sq->prod;
	spin_unlock_irqrestore(&sq->lock, flags);

	index = prod & sq->size_mask;
	mlx4_fc_q_info_get(sq, index, struct sk_buff *) = skb;

	tx_desc = sq->buf + index * RFCI_SQ_BB_SIZE;

	tx_desc->ctrl.size = cpu_to_be16((desc_size / 16) & 0x3f);
	if (vhba->fc_vlan_id != -1) {
		tx_desc->ctrl.size |= cpu_to_be16(MLX4_FC_BIT_INS_VLAN);
		tx_desc->ctrl.vlan =
		    cpu_to_be16(vhba->fc_vlan_id | vhba->fc_vlan_prio << 13);
	}

	tx_desc->ctrl.flags = cpu_to_be32(MLX4_FC_BIT_TX_COMP |
					  MLX4_FC_BIT_NO_ICRC |
					  MLX4_FC_BIT_TX_FCRC_CS);

	op_own = cpu_to_be32(MLX4_FC_RFCI_OP_SEND) |
			     cpu_to_be32((u32) fceof << 16) |
			     ((prod & sq->size) ?
			      cpu_to_be32(MLX4_FC_BIT_DESC_OWN) : 0);
	data = &tx_desc->data;

	dma = pci_map_single(mfcdev->dev->pdev, skb->data, skb->len - 5,
			     PCI_DMA_TODEVICE);
	data->addr = cpu_to_be64(dma);
	data->count = cpu_to_be32(skb->len - 5);
	data->mem_type = cpu_to_be32(mfcdev->mr.key);	/* always snoop */

	/* Ensure new descirptor (and ownership of next descirptor) hits memory
	 * before setting ownership of this descriptor to HW */
	wmb();
	tx_desc->ctrl.op_own = op_own;

	/* Ring doorbell! */
	wmb();
	writel(rfci->fc_qp.doorbell_qpn, mfcdev->uar_map + MLX4_SEND_DOORBELL);

	return 0;
}

int mlx4_start_rfci_data(struct mlx4_fc_vhba *vhba, u64 mac)
{
	int err = 0;

	/*
	 * Remove any previously-set unicast MAC filter.
	 * Add secondary FCoE MAC address filter for our OUI.
	 */
	if (vhba->rfci[RFCI_DATA].fc_mac_idx != -1) {
		FCLOG(fc_log_ctrl, "Destroying old data RFCI\n");
		mlx4_fc_destroy_rfci(vhba, &vhba->rfci[RFCI_DATA]);
	}

	err = mlx4_fc_create_rfci(vhba, &vhba->rfci[RFCI_DATA], mac);
	if (err) {
		mlx4_err(vhba->fport->mfcdev->dev,
			 "Could not create data RFCI QP, err=%d\n", err);
		return err;
	}
	FCLOG(fc_log_ctrl, "Created RFCI for data MAC\n");
	err = mlx4_fc_init_rfci(vhba, &vhba->rfci[RFCI_DATA]);
	if (err) {
		mlx4_err(vhba->fport->mfcdev->dev,
			 "Could not init data RFCI QP, err=%d\n", err);
		return err;
	}
	FCLOG(fc_log_ctrl, "RFCI in state RTS\n");
	return err;
}

static void mlx4_fc_recv_flogi(struct fc_lport *lp, struct fc_frame *fp,
			       u8 sa[6])
{
	struct mlx4_fc_vhba *vhba = lport_priv(lp);
	struct fc_frame_header *fh;
	u8 op;

	fh = fc_frame_header_get(fp);
	if (fh->fh_type != FC_TYPE_ELS)
		return;
	op = fc_frame_payload_op(fp);
	if (op == ELS_LS_ACC && fh->fh_r_ctl == FC_RCTL_ELS_REP &&
	    vhba->flogi_oxid == ntohs(fh->fh_ox_id)) {
		memcpy(vhba->dest_addr, sa, ETH_ALEN);
		memcpy(vhba->my_fid, fh->fh_d_id, 3);
		fc_fcoe_set_mac(vhba->rfci[RFCI_DATA].mac, fh->fh_d_id);
		mlx4_fc_flogi_finished(lp);
		vhba->flogi_progress = 0;
	} else if (op == ELS_FLOGI && fh->fh_r_ctl == FC_RCTL_ELS_REQ && sa) {
		/* TODO: support for p2p */
		memcpy(vhba->dest_addr, sa, ETH_ALEN);
	}
}

static void mlx4_fc_rx_rfci(struct work_struct *work)
{
	struct mlx4_fc_rfci_rx_info *fr =
	    container_of(work, struct mlx4_fc_rfci_rx_info, work);
	u_int32_t fr_len;
	u_int hlen;
	u_int tlen;
	struct mlx4_fc_vhba *vhba = fr->vhba;
	struct fc_lport *lp = vhba->lp;
	struct fcoe_dev_stats *stats = lp->dev_stats[smp_processor_id()];
	struct fc_frame_header *fh;
	struct sk_buff *skb = fr->skb;
	struct fcoe_crc_eof_old *cp;
	enum fc_sof sof;
	struct fc_frame *fp;
	u8 mac[6] = { 0 };
	struct fcoe_hdr_old *fchp;
	u_int len;

	/*
	 * Save source MAC address before discarding header.
	 */
	if (unlikely(vhba->flogi_progress))
		memcpy(mac, eth_hdr(skb)->h_source, ETH_ALEN);

	/*
	 * Check the header and pull it off.
	 */
	hlen = vhba->fcoe_hlen;
	ASSERT(hlen == sizeof(*fchp));
	fchp = (struct fcoe_hdr_old *)skb->data;
	len = ntohs(fchp->fcoe_plen);
	skb_pull(skb, sizeof(*fchp));
	fr_len = FCOE_DECAPS_LEN(len);
	fr_len = fr_len * FCOE_WORD_TO_BYTE;
	if (unlikely(fr_len < sizeof(struct fc_frame_header) +
		     sizeof(cp->fcoe_crc32))) {
		if (stats->ErrorFrames < 5)
			FCPRINT("length error: len_sof %x\n", len);
		stats->ErrorFrames++;
		goto bad_packet;
	}
	sof = FCOE_DECAPS_SOF(len);
	fr_len -= sizeof(cp->fcoe_crc32);
	tlen = sizeof(struct fcoe_crc_eof_old);

	if (skb_is_nonlinear(skb))
		skb_linearize(skb);	/* not ideal */
	if (unlikely(fr_len + tlen > skb->len)) {
		if (stats->ErrorFrames < 5)
			FCPRINT("length error fr_len 0x%x skb->len 0x%x\n",
				fr_len + tlen, skb->len);
		stats->ErrorFrames++;
		goto bad_packet;
	}
	stats->RxFrames++;
	stats->RxWords += fr_len / FCOE_WORD_TO_BYTE;

	fp = (struct fc_frame *)skb;
	fc_frame_init(fp);
	cp = (struct fcoe_crc_eof_old *)(skb->data + fr_len);
	fr_eof(fp) = cp->fcoe_eof;
	fr_sof(fp) = sof;
	fr_dev(fp) = lp;

	fh = fc_frame_header_get(fp);
	ASSERT(fh);
#if FC_DEBUG
	if (fc_log_ctrl <= LOG_LEVEL(mlx4_fc_log_level)) {
		FCLOG(fc_log_ctrl,
		      "Received frame. ox_id: 0x%x fh_r_ctl: 0x%x\n",
		      ntohs(fh->fh_ox_id), fh->fh_r_ctl);
	}
#endif

	if (unlikely(vhba->flogi_progress))
		mlx4_fc_recv_flogi(lp, fp, mac);

	fc_exch_recv(lp, lp->emp, fp);

	/* no need for kfree_skb() - skb was already freed inside 
	   fc_exch_recv() */
	return;

bad_packet:
	kfree_skb(skb);
}

int mlx4_fc_send(struct fc_lport *lp, struct fc_frame *fp)
{
	struct mlx4_fc_vhba *vhba = lport_priv(lp);
	struct fc_frame_header *fh;
	struct sk_buff *skb;
	uint8_t sof, eof;
	u_int hlen;
	u_int tlen;
	int wlen;
	struct ethhdr *eh;
	struct fcoe_crc_eof *cp;
	int flogi_in_progress = 0;
	struct fcoe_hdr_old *ohp;
	int data_channel;

	ASSERT((fr_len(fp) % sizeof(uint32_t)) == 0);

	fh = fc_frame_header_get(fp);

	if (unlikely(fh->fh_r_ctl == FC_RCTL_ELS_REQ)) {
		if (fc_frame_payload_op(fp) == ELS_FLOGI) {
			vhba->flogi_oxid = ntohs(fh->fh_ox_id);
			vhba->flogi_progress = 1;
			flogi_in_progress = 1;
			vhba->rfci_rx_enabled = 1;
			if (mlx4_debug_mode == 0) {
				u8 flogi_did[] = { 0xff, 0xff, 0xfe };
				fc_fcoe_set_mac(vhba->dest_addr, flogi_did);
			}
		}
	}

	if (flogi_in_progress || (mlx4_debug_mode == 1))
		data_channel = RFCI_CTRL;
	else
		data_channel = RFCI_DATA;

	skb = fp_skb(fp);
	sof = fr_sof(fp);
	eof = fr_eof(fp);

	hlen = sizeof(struct fcoe_hdr_old);
	tlen = sizeof(struct fcoe_crc_eof_old);

	ASSERT(!skb_is_nonlinear(skb));
	cp = (struct fcoe_crc_eof *)skb_put(skb, tlen);

	wlen = (skb->len - tlen + sizeof(u32)) / FCOE_WORD_TO_BYTE;

	eh = (struct ethhdr *)skb_push(skb, hlen + sizeof(struct ethhdr));

	memcpy(eh->h_dest, vhba->dest_addr, ETH_ALEN);
	memcpy(eh->h_source, vhba->rfci[data_channel].mac, ETH_ALEN);

	eh->h_proto = __constant_htons(ETH_P_FCOE);

	ASSERT(hlen == sizeof(*ohp));
	ohp = (struct fcoe_hdr_old *)(eh + 1);
	ohp->fcoe_plen = htons(FCOE_ENCAPS_LEN_SOF(wlen, sof));

	fr_dev(fp) = lp;

	FCLOG(fc_log_ctrl, "Send frame. ox_id: 0x%x fh_r_ctl: 0x%x\n",
	      ntohs(fh->fh_ox_id), fh->fh_r_ctl);

	return mlx4_do_rfci_xmit(vhba, data_channel, skb, eof);
}
