/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/etherdevice.h>
#include <scsi/scsi.h>
#include <scsi/libfc.h>
#include <scsi/fc/fc_fcoe.h>

#include "mlx4_fc.h"

/* scsi stuff */
void mlx4_fc_rx_scsi(struct mlx4_fc_vhba *vhba, struct fcp_resp *fc_rp,
		     struct scsi_cmnd *sc_cmd, size_t xfer_len,
		     struct mlx4_fmr *fmr, u32 xno)
{
	enum dma_data_direction data_dir;
	u_int8_t cdb_op;
	struct fcp_resp_ext *rp_ex;
	struct fcp_resp_rsp_info *fc_rp_info;
	u_int32_t respl = 0;
	u_int32_t snsl = 0;
	u32 scsi_resid;
	u8 cdb_status;
	unsigned data_len = scsi_bufflen(sc_cmd);

	/* things from openfc_scsi_rcv(), RESPONSE branch */
	cdb_status = fc_rp->fr_status;
	if ((fc_rp->fr_flags == 0) && (fc_rp->fr_status == 0)) {
		data_dir = sc_cmd->sc_data_direction;
		cdb_op = sc_cmd->cmnd[0];
		if (data_dir == DMA_FROM_DEVICE) {
			if (data_len > xfer_len) {
				if ((cdb_op != READ_10 ||
				     cdb_op != READ_6 ||
				     cdb_op != WRITE_10 || cdb_op != WRITE_6)) {
					sc_cmd->result = DID_IMM_RETRY << 16;
					FCPRINT("data_len (%x) > xfer_len (%zx)"
						", and not READ/WRITE-10/6:"
						" SUGGEST_RETRY\n",
						data_len, xfer_len);
				} else {
					FCPRINT("xfer len less then "
						"data_len: xfer_len %zx"
						" data_len %x, OPENFC_DATA_UNDRUN \n",
						xfer_len, data_len);
					if (cdb_status == 0) {
						sc_cmd->result =
						    (DID_OK << 16) | cdb_status;
					} else {
						CMD_RESID_LEN(sc_cmd) =
						    scsi_resid;
						sc_cmd->result =
						    (DID_ERROR << 16) |
						    cdb_status;
					}
				}
			} else if (data_len < xfer_len) {
				if ((cdb_op != READ_10 ||
				     cdb_op != READ_6 ||
				     cdb_op != WRITE_10 || cdb_op != WRITE_6)) {
					sc_cmd->result = DID_IMM_RETRY << 16;
					FCPRINT("data_len (%x) < xfer_len (%zx)"
						", and not READ/WRITE-10/6:"
						" SUGGEST_RETRY\n",
						data_len, xfer_len);
				} else {
					FCPRINT("xfer len greater than "
						"data_len xfer_len %zx"
						" data_len %x, OPENFC_DATA_OVRRUN \n",
						xfer_len, data_len);
					sc_cmd->result =
					    (DID_ERROR << 16) | cdb_status;
				}
			}
		}
	} else {
		rp_ex = (void *)(fc_rp + 1);
		fc_rp_info = (struct fcp_resp_rsp_info *)(rp_ex + 1);
		if (fc_rp->fr_flags & FCP_RSP_LEN_VAL) {
			respl = ntohl(rp_ex->fr_rsp_len);
			if ((respl != 0 && respl != 4 && respl != 8) ||
			    (fc_rp_info->rsp_code != FCP_TMF_CMPL)) {
				sc_cmd->result = (DID_ERROR << 16);
				FCPRINT("bad respl: %d, and rsp_code isn't"
					" FCP_TMF_CMPL... OPENFC_ERROR\n",
					respl);
			}
		}
		if (fc_rp->fr_flags & FCP_SNS_LEN_VAL) {
			snsl = ntohl(rp_ex->fr_sns_len);
			if (snsl > SCSI_SENSE_BUFFERSIZE)
				snsl = SCSI_SENSE_BUFFERSIZE;
			memcpy(sc_cmd->sense_buffer,
			       &fc_rp_info->_fr_resvd[0] + respl, snsl);
			FCLOG(fc_log_data, "Copied SENSE data\n");
		}
		if (fc_rp->fr_flags & FCP_RESID_UNDER) {
			scsi_resid = ntohl(rp_ex->fr_resid);
			/*
			 * The cmnd->underflow is the minimum number of
			 * bytes that must be transfered for this
			 * command.  Provided a sense condition is not
			 * present, make sure the actual amount
			 * transferred is at least the underflow value
			 * or fail.
			 */
			if (!(fc_rp->fr_flags & FCP_SNS_LEN_VAL) &&
			    (fc_rp->fr_status == 0) &&
			    (scsi_bufflen(sc_cmd) -
			     scsi_resid) < sc_cmd->underflow) {
				sc_cmd->result = (DID_ERROR << 16);
				FCPRINT("OPENFC_ERROR - FCP_RESID_UNDER...\n");
			}
		} else if (fc_rp->fr_flags & FCP_RESID_OVER) {
			sc_cmd->result = (DID_ERROR << 16);
			FCPRINT("OPENFC_ERROR - FCP_RESID_OVER\n");
		}
	}
	/* ... done */

	FCLOG(fc_log_data, "calling done. sc_cmd: %p, sc_cmd->result = 0x%x\n",
	      sc_cmd, sc_cmd->result);

	sc_cmd->scsi_done(sc_cmd);
}
